#!/usr/bin/env python3
"""
UI-DISPLAY.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A demo app that displays all of the available UI actions
and their associated images, and demonstrates how event
handlers are defined.
"""

import sys

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication
from plib.ui.widgets import *


class UIDisplayTester(PApplication):
    
    about_data = {
        'name': "UIDisplayTester",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "UI Display Demo",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net"
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    menu_actions = [
        (MENU_FILE, (
            ACTION_FILE_NEW, ACTION_FILE_OPEN, ACTION_FILE_SAVE, ACTION_FILE_SAVEAS,
            ACTION_FILE_CLOSE, ACTION_EXIT,
        )),
        (MENU_EDIT, (
            ACTION_EDIT_UNDO, ACTION_EDIT_REDO,
            ACTION_EDIT_CUT, ACTION_EDIT_COPY, ACTION_EDIT_PASTE,
            ACTION_EDIT_DELETE, ACTION_EDIT_SELECTALL, ACTION_EDIT_SELECTNONE, ACTION_EDIT_CLEAR,
        )),
        (MENU_ACTION, (
            ACTION_VIEW, ACTION_EDIT, ACTION_APPLY, ACTION_REFRESH,
            ACTION_OK, ACTION_CANCEL,
            ACTION_ADD, ACTION_REMOVE,
            ACTION_COMMIT, ACTION_ROLLBACK,
        )),
        (MENU_OPTIONS, (
            ACTION_PREFS,
        )),
        (MENU_HELP, (
            ACTION_ABOUT, ACTION_ABOUT_TOOLKIT,
        )),
    ]
    
    toolbar_actions = [
        (ACTION_FILE_NEW, ACTION_FILE_OPEN, ACTION_FILE_SAVE, ACTION_FILE_SAVEAS, ACTION_FILE_CLOSE,),
        (ACTION_EDIT_UNDO, ACTION_EDIT_REDO,),
        (ACTION_EDIT_CUT, ACTION_EDIT_COPY, ACTION_EDIT_PASTE,),
        (ACTION_EDIT_DELETE, ACTION_EDIT_SELECTALL, ACTION_EDIT_SELECTNONE, ACTION_EDIT_CLEAR,),
        (ACTION_VIEW, ACTION_EDIT, ACTION_APPLY, ACTION_REFRESH,),
        (ACTION_OK, ACTION_CANCEL,),
        (ACTION_ADD, ACTION_REMOVE,),
        (ACTION_COMMIT, ACTION_ROLLBACK,),
        (ACTION_PREFS, ACTION_ABOUT, ACTION_ABOUT_TOOLKIT, ACTION_EXIT,),
    ]
    
    status_labels = [
        ('action', "", PANEL_SUNKEN),
    ]
    
    main_title = "UI Display Demo"
    
    main_size = SIZE_MAXIMIZED
    
    main_widget = text('output', "PLIB3 UI Display Demo Application.",
                       font=("Courier New", 12))
    
    query_on_exit = False
    
    def set_status_text(self, text):
        print(text)
        self.main_window.statusbar.set_text(text)
        self.label_action.caption = text
    
    def on_file_new(self):
        self.set_status_text("ACTION_FILE_NEW")
    
    def on_file_open(self):
        self.set_status_text("ACTION_FILE_OPEN")
    
    def on_file_save(self):
        self.set_status_text("ACTION_FILE_SAVE")
    
    def on_file_saveas(self):
        self.set_status_text("ACTION_FILE_SAVEAS")
    
    def on_file_close(self):
        self.set_status_text("ACTION_FILE_CLOSE")
    
    def on_edit_undo(self):
        self.set_status_text("ACTION_EDIT_UNDO")
    
    def on_edit_redo(self):
        self.set_status_text("ACTION_EDIT_REDO")
    
    def on_edit_cut(self):
        self.set_status_text("ACTION_EDIT_CUT")
    
    def on_edit_copy(self):
        self.set_status_text("ACTION_EDIT_COPY")
    
    def on_edit_paste(self):
        self.set_status_text("ACTION_EDIT_PASTE")
    
    def on_edit_delete(self):
        self.set_status_text("ACTION_EDIT_DELETE")
    
    def on_edit_selectall(self):
        self.set_status_text("ACTION_EDIT_SELECTALL")
    
    def on_edit_selectnone(self):
        self.set_status_text("ACTION_EDIT_SELECTNONE")
    
    def on_edit_clear(self):
        self.set_status_text("ACTION_EDIT_CLEAR")
    
    def on_view(self):
        self.set_status_text("ACTION_VIEW")
    
    def on_edit(self):
        self.set_status_text("ACTION_EDIT")
    
    def on_refresh(self):
        self.set_status_text("ACTION_REFRESH")
    
    def on_add(self):
        self.set_status_text("ACTION_ADD")
    
    def on_remove(self):
        self.set_status_text("ACTION_REMOVE")
    
    def on_apply(self):
        self.set_status_text("ACTION_APPLY")
    
    def on_commit(self):
        self.set_status_text("ACTION_COMMIT")
    
    def on_rollback(self):
        self.set_status_text("ACTION_ROLLBACK")
    
    def on_ok(self):
        self.set_status_text("ACTION_OK")
    
    def on_cancel(self):
        self.set_status_text("ACTION_CANCEL")
    
    def on_prefs(self):
        self.set_status_text("ACTION_PREFS")
    
    def on_about(self):
        self.set_status_text("ACTION_ABOUT")
        self.about()
    
    def on_about_toolkit(self):
        self.set_status_text("ACTION_ABOUT_TOOLKIT")
        self.about_toolkit()
    
    def on_exit(self):
        self.set_status_text("ACTION_EXIT")
        self.exit_app()
    
    def accept_close(self):
        return (not self.query_on_exit) or self.message_box.query2(
            "Application Exit",
            "Exit {}?".format(self.about_data['name'])
        ) == ANSWER_OK


if __name__ == "__main__":
    from plib.stdlib.options import parse_options
    optlist = (
        ("-l", "--large-icons", {
            'action': "store_true",
            'dest': "large_icons", 'default': False,
            'help': "Use large toolbar icons"
        }),
        ("-s", "--show-labels", {
            'action': "store_true",
            'dest': "show_labels", 'default': False,
            'help': "Show toolbar button labels"
        }),
        ("-q", "--query-on-exit", {
            'action': "store_true",
            'dest': "query_on_exit", 'default': False,
            'help': "Ask for confirmation on app exit"
        }),
    )
    opts, args = parse_options(optlist)
    # The options object supports a dictionary interface,
    # making it easy to update class fields from it
    for opt, value in opts.items():
        setattr(UIDisplayTester, opt, value)
    UIDisplayTester().run()



#!/usr/bin/env python3
"""
Module QT5.MAINWIN -- Python Qt 5 Main Window Objects
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI main window objects. It is kept
separate from the ``app`` module so that applications that do not
require a full main window but just a "bare" top window do not have
to import any code that is only used with a full main window.
"""

from PyQt5 import QtCore as qtc, QtGui as qtg, QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.mainwin import PMenuBase, PToolBarBase, PStatusLabelBase, PStatusBarBase, PActionBase, PMainWindowBase

from .app import PQtWidgetMeta, PQtSignalBase, PQtActionMixin, PQtWidgetBase, PTopWindow
from .form import PPanel
from .label import PTextLabel


class PQtMenu(qt.QMenuBar, PQtWidgetBase, PMenuBase):
    """Customized Qt menu class.
    """
    
    widget_class = qt.QMenuBar
    
    popup_class = qt.QMenu
    
    def __init__(self, main_window):
        qt.QMenuBar.__init__(self, main_window)
        PMenuBase.__init__(self, main_window)
        main_window.setMenuBar(self)
    
    def do_add_popup(self, title, popup):
        popup.setTitle(title)
        self.addMenu(popup)
    
    def add_popup_action(self, action, popup):
        popup.addAction(action)


class PQtToolBar(qt.QToolBar, PQtWidgetBase, PToolBarBase):
    """Customized Qt toolbar class.
    """
    
    widget_class = qt.QToolBar
    
    def __init__(self, main_window):
        qt.QToolBar.__init__(self, main_window)
        if main_window.app.large_icons:
            sz = self.iconSize()
            self.setIconSize(qtc.QSize(
                *[i * 3 / 2 for i in (sz.width(), sz.height())]))
        if main_window.app.show_labels:
            style = qtc.Qt.ToolButtonTextUnderIcon
        else:
            style = qtc.Qt.ToolButtonIconOnly
        self.setToolButtonStyle(style)
        PToolBarBase.__init__(self, main_window)
        main_window.addToolBar(self)
    
    def add_action(self, action):
        self.addAction(action)
    
    def add_separator(self):
        self.addSeparator()


class PQtStatusPanel(PPanel):
    
    label = None
    
    def __init__(self, manager, parent, caption="", style=PANEL_NONE):
        PPanel.__init__(self, manager, parent, ALIGN_RIGHT, LAYOUT_HORIZONTAL,
                        style=style, margin=PANEL_MARGIN, spacing=PANEL_SPACING)
        self.label = PTextLabel(manager, self, caption)
        self.add_widget(self.label)
        self.do_layout()
    
    def create_contents(self, contents):
        assert not contents  # we will create our label manually, not using specs
    
    def do_layout(self):
        # This ensures that we only do the layout
        if self.label:
            super(PQtStatusPanel, self).do_layout()


class PStatusLabel(PStatusLabelBase):
    
    def init_status_label(self):
        self.panel = PQtStatusPanel(self._manager, self._parent, caption=self._caption, style=self._style)
        self.label = self.panel.label
        self.actual_widget = self.panel
        
        self.text = self.label.text
        self.setText = self.label.setText
    
    fn_get_text = 'text'
    fn_set_text = 'setText'


class PQtStatusBar(qt.QStatusBar, PQtWidgetBase, PStatusBarBase):
    
    widget_class = qt.QStatusBar
    
    text_area_class = PTextLabel
    
    def __init__(self, main_window):
        qt.QStatusBar.__init__(self, main_window)
        PStatusBarBase.__init__(self, main_window)
        main_window.setStatusBar(self)
    
    def add_widget(self, widget, expand=False, custom=True):
        if custom:
            self.addPermanentWidget(widget, int(expand))
        else:
            self.addWidget(widget, int(expand))


class PQtAction(qt.QAction, PQtActionMixin, PQtSignalBase, PActionBase):
    """Customized Qt action class.
    """
    
    def __init__(self, key, main_window):
        qt.QAction.__init__(self, main_window)
        self.setIcon(self.get_icon(key))
        self.setText(self.get_menu_str(key))
        self.setToolTip(self.get_toolbar_str(key))
        self.setStatusTip(self.get_statusbar_str(key))
        s = self.get_accel_str(key)
        if s is not None:
            if isinstance(s, str):
                self.setShortcut(qtg.QKeySequence(s))
            elif isinstance(s, qtg.QKeySequence.StandardKey):
                self.setShortcuts(qtg.QKeySequence.keyBindings(s))
        PActionBase.__init__(self, key, main_window)
    
    def enable(self, enabled):
        self.setEnabled(enabled)


class PMainWindow(PTopWindow, PMainWindowBase,
                  metaclass=PQtWidgetMeta):
    """Customized Qt main window class.
    """
    
    menu_class = PQtMenu
    toolbar_class = PQtToolBar
    statusbar_class = PQtStatusBar
    action_class = PQtAction



#!/usr/bin/env python3
"""
UI-SIGNALS.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A demo app that tests all of the available UI signals.
"""

import sys
import os

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication
from plib.ui.output import PTextOutput
from plib.ui.widgets import *


number_strings = ("One", "Two", "Three", "Four", "Five", "Six")


class UISignalTester(PApplication):
    
    about_data = {
        'name': "UISignalTester",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "UI Signal Test Demo",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net"
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    main_title = "UI Signal Tester"
    
    main_size = SIZE_CLIENTWRAP
    main_placement = MOVE_CENTER
    
    main_widget = frame(ALIGN_JUST, LAYOUT_HORIZONTAL, [
        tabwidget('panels', [
            ('Dialog', tab(ALIGN_JUST, LAYOUT_VERTICAL, [
                button('action', "Test Clicked"),
                checkbox('option', "Test Toggled"),
                sorted_combo('selection', (numstr.lower() for numstr in number_strings)),
                edit('text'),
                num_edit('num'),
                sorted_listbox('list', ("Item {}".format(numstr) for numstr in number_strings)),
                padding(),
            ])),
            ('Memo', memo('notes')),
            ('Tree', treeview('tree', [("Title", WIDTH_CONTENTS), ("Description", WIDTH_STRETCH)], (
                (("Item {}".format(numstr), "Tree item {}".format(numstr.lower())), tuple(
                    (("Sub-item {} {}".format(numstr, substr), "Tree sub-item {} {}".format(numstr, substr)), ())
                    for substr in number_strings[:3]
                ))
                for numstr in number_strings
            ), auto_expand=True)),
            ('List', sorted_listview('items', [("Title", WIDTH_CONTENTS), ("Description", WIDTH_STRETCH)], (
                ("Item {}".format(numstr), "List item {}".format(numstr.lower()))
                for numstr in number_strings
            ))),
            ('Table', table('cells', [("Title", WIDTH_CONTENTS), ("Description", WIDTH_STRETCH)], (
                ("Row {}".format(numstr), "Table row {}".format(numstr.lower()))
                for numstr in number_strings
            ))),
        ]),
        text('output'),
    ])
    
    def before_create(self):
        # We can't call output_message here because widgets aren't available yet
        print("before_create")
    
    def after_create(self):
        self.outputfile = PTextOutput(self.text_output)
        self.output_message("after_create")
    
    def output_message(self, message):
        print(message)
        self.outputfile.write("{}{}".format(message, os.linesep))
        self.outputfile.flush()
    
    def focus_in(self, widget_name):
        self.output_message("SIGNAL_FOCUS_IN {}".format(widget_name))
    
    def focus_out(self, widget_name):
        self.output_message("SIGNAL_FOCUS_OUT {}".format(widget_name))
    
    def on_text_focus_in(self):
        self.focus_in('edit_text')
    
    def on_text_focus_out(self):
        self.focus_out('edit_text')
    
    def on_notes_focus_in(self):
        self.focus_in('memo_notes')
    
    def on_notes_focus_out(self):
        self.focus_out('memo_notes')
    
    def on_panels_selected(self, index):
        assert self.tabwidget_panels.current_index() == index
        self.output_message("SIGNAL_TABSELECTED {} {}".format(
            index, self.tabwidget_panels[index][0])
        )
    
    def on_action_clicked(self):
        self.output_message("SIGNAL_CLICKED")
    
    def on_option_toggled(self, checked):
        assert checked == self.checkbox_option.checked
        self.output_message("SIGNAL_TOGGLED {}".format(('off', 'on')[checked]))
    
    def on_selection_selected(self, index):
        assert self.combo_selection.current_index() == index
        assert self.combo_selection.current_text() == self.combo_selection[index]
        self.output_message("SIGNAL_SELECTED {} {}".format(index, self.combo_selection.current_text()))
    
    def on_text_changed(self, text):
        assert text == self.edit_text.edit_text
        self.output_message("SIGNAL_EDITCHANGED {}".format(text))
    
    def on_text_enter(self):
        self.output_message("SIGNAL_ENTER")
    
    def on_num_changed(self, value):
        assert value == self.edit_num.edit_value
        self.output_message("SIGNAL_EDITCHANGED {}".format(value))
    
    def on_list_selected(self, item):
        assert item == self.listbox_list.current_text()
        assert item == self.listbox_list[self.listbox_list.current_index()]
        self.output_message("SIGNAL_LISTSELECTED {} {} {}".format(
            'listbox', self.listbox_list.current_index(), item
        ))
    
    def on_notes_changed(self):
        self.output_message("SIGNAL_TEXTCHANGED {}".format(self.memo_notes.edit_text))
    
    def on_notes_mod_changed(self, changed):
        self.output_message("SIGNAL_TEXTMODCHANGED {} {}".format(self.memo_notes.edit_text, changed))
    
    def on_notes_state_changed(self):
        self.output_message("SIGNAL_TEXTSTATECHANGED memo_notes")
    
    def on_tree_selected(self, cols, children):
        assert (cols, children) == (self.treeview_tree.current_item().cols, self.treeview_tree.current_item().children)
        self.output_message("SIGNAL_LISTSELECTED {} ({}) {} {} {}".format(
            'treeview', " ".join(str(i) for i in self.treeview_tree.current_indexes()), len(children), *cols
        ))
    
    def on_items_selected(self, cols):
        assert cols == self.listview_items.current_item().cols
        assert cols == self.listview_items[self.listview_items.current_index()]
        self.output_message("SIGNAL_LISTSELECTED {} {} {} {}".format(
            'listview', self.listview_items.current_index(), *cols
        ))
    
    def on_cells_selected(self, curr_row, curr_col, prev_row, prev_col):
        assert self.table_cells.current_row() == curr_row
        assert self.table_cells.current_col() == curr_col
        assert self.table_cells.current_cell() == self.table_cells[curr_row][curr_col]
        self.output_message("SIGNAL_CELLSELECTED ({}, {}) -> ({}, {}) {}".format(
            str(prev_row), str(prev_col), str(curr_row), str(curr_col), self.table_cells.current_cell())
        )
    
    def on_cells_changed(self, row, col):
        assert self.table_cells.current_row() == row
        assert self.table_cells.current_col() == col
        assert self.table_cells.current_cell() == self.table_cells[row][col]
        self.output_message("SIGNAL_CELLCHANGED {} {} {}".format(
            str(row), str(col), self.table_cells.current_cell())
        )
    
    queryclose_handled = False
    
    def on_app_queryclose(self):
        self.queryclose_handled = True
        self.output_message("SIGNAL_QUERYCLOSE")
    
    def accept_close(self):
        print("SIGNAL_QUERYCLOSE handled:", self.queryclose_handled)
        return self.queryclose_handled
    
    def on_app_closing(self):
        self.output_message("SIGNAL_CLOSING")
    
    def on_app_shown(self):
        self.output_message("SIGNAL_SHOWN")
    
    def on_app_hidden(self):
        self.output_message("SIGNAL_HIDDEN")
    
    def before_quit(self):
        # We can't call output_message here because widgets might not be available
        # any more
        print("before_quit")


if __name__ == "__main__":
    UISignalTester().run()



#!/usr/bin/env python3
"""
Module PYSIDE2.COMBO -- Python PySide 2 Combo Box Widgets
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for combo boxes.
"""

from PySide2 import QtWidgets as qt

from plib.ui.base.combo import PComboBoxBase, PNumComboBoxBase, PSortedComboBoxBase

from .app import PQtSequenceWidget


class PQtComboBoxBase(PQtSequenceWidget, qt.QComboBox):
    
    fix_width_on_resize = True
    
    def __init__(self, parent):
        qt.QComboBox.__init__(self, parent)
        self.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Fixed)
    
    def current_text(self):
        return str(self.currentText())
    
    # Note that there's no quick override for set_current_text in Qt; the
    # corresponding method to the above doesn't do what we want (it changes
    # the stored text in the combo instead of selecting the text we give it)
    
    def current_index(self):
        return self.currentIndex()
    
    def set_current_index(self, index):
        self.setCurrentIndex(index)
    
    def count(self, value):
        # Method name collision, we want it to be the Python sequence
        # count method here.
        return PComboBoxBase.count(self, value)
    
    def _indexlen(self):
        # Let this method access the Qt combo box count method.
        return qt.QComboBox.count(self)
    
    def _get_data(self, index):
        return str(self.itemText(index))
    
    def _set_data(self, index, value):
        self.setItemText(index, str(value))
    
    def _add_data(self, index, value):
        self.insertItem(index, str(value))
    
    def _del_data(self, index):
        self.removeItem(index)


class PComboBox(PQtComboBoxBase, PComboBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None):
        
        PQtComboBoxBase.__init__(self, parent)
        PComboBoxBase.__init__(self, manager, parent, items, value=value,
                               geometry=geometry)


class PNumComboBox(PQtComboBoxBase, PNumComboBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None):
        
        PQtComboBoxBase.__init__(self, parent)
        PNumComboBoxBase.__init__(self, manager, parent, items, value=value,
                                  geometry=geometry)


class PSortedComboBox(PQtComboBoxBase, PSortedComboBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, key=None):
        
        PQtComboBoxBase.__init__(self, parent)
        PSortedComboBoxBase.__init__(self, manager, parent, items, value=value,
                                     geometry=geometry, key=key)



#!/usr/bin/env python3
"""
Module WX.DISPLAY -- Python wxWidgets Text Display Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for text display widgets.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.display import PTextDisplayBase

from .app import PWxWidget, ec_scroll_style


class PTextDisplay(PWxWidget, wx.TextCtrl, PTextDisplayBase):
    
    _style = wx.TE_MULTILINE | wx.TE_PROCESS_TAB | wx.TE_READONLY
    
    _align = True  # used by panel to determine placement
    _expand = True
    
    def __init__(self, manager, parent, text=None,
                 scrolling=False, font=None):
        
        if scrolling:
            self._style = self._style | ec_scroll_style
        wx.TextCtrl.__init__(self, parent, style=self._style)
        PTextDisplayBase.__init__(self, manager, parent, text,
                                  scrolling=scrolling, font=font)
    
    def setup_scrolling(self, scrolling):
        pass  # this must be done in the constructor in wxWidgets
    
    def get_text(self):
        return self.GetValue()
    
    def set_text(self, value):
        self.SetValue(value)



#!/usr/bin/env python3
"""
Module WX.EDITCTRL -- Python wxWidgets Editing Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for edit controls.
"""

import wx
from wx.lib.evtmgr import eventManager

from plib.ui.defs import *
from plib.ui.base.editctrl import PEditBoxBase, PNumEditBoxBase, PEditControlBase

from .app import PWxWidget, ec_scroll_style


class PWxEditBase(PWxWidget, wx.TextCtrl):
    
    fn_get_text = 'GetValue'
    fn_set_text = 'SetValue'
    
    fn_get_readonly = 'NotEditable'
    fn_set_readonly = 'SetNotEditable'
    
    _align = False  # used by panel to determine placement
    
    def __init__(self, parent, expand):
        self._expand_horiz = expand
        wx.TextCtrl.__init__(self, parent, style=self._style)
    
    def setup_expand(self, expand):
        pass  # this is handled differently in wx, see constructor above
    
    def NotEditable(self):
        return not self.IsEditable()
    
    def SetNotEditable(self, value):
        self.SetEditable(not value)


class PEditBox(PWxEditBase, PEditBoxBase):
    
    _style = wx.TE_PROCESS_ENTER
    
    def __init__(self, manager, parent, text="",
                 geometry=None, expand=True):
        
        PWxEditBase.__init__(self, parent, expand)
        PEditBoxBase.__init__(self, manager, parent, text,
                              geometry=geometry, expand=expand)


# FIXME: Wx docs say this class should exist, but can't find it

class IntegerValidator(wx.Validator):
    
    def __init__(self, value=0):
        wx.Validator.__init__(self)
        self.value = value
    
    # FIXME: Wx docs say you're supposed to always implement Clone, but
    # doing that breaks the validator
    
    def IsValid(self, value):
        try:
            int(value)
        except (TypeError, ValueError):
            return False
        else:
            return True
    
    def OnTextEvent(self, event):
        text = self.GetWindow().GetValue()
        if not self.IsValid(text):
            self.GetWindow().ChangeValue(str(self.value))
            event.StopPropagation()
        else:
            self.value = int(text)
            event.Skip()
    
    def Validate(self, parent):
        return IsValid(parent.GetValue())
    
    def TransferFromWindow(self):
        self.value = int(self.GetWindow().GetValue())
    
    def TransferToWindow(self):
        self.GetWindow().SetValue(str(self.value))


class PNumEditBox(PWxEditBase, PNumEditBoxBase):
    
    _style = wx.TE_PROCESS_ENTER
    
    def __init__(self, manager, parent, value=0,
                 geometry=None, expand=True):
        
        PWxEditBase.__init__(self, parent, expand)
        self.val = val = IntegerValidator()
        self.SetValidator(val)
        val.SetWindow(self)  # FIXME: isn't Wx supposed to do this automatically in SetValidator?
        # Bypass the normal setup_notify mechanism because we don't want a wrapper
        eventManager.Register(val.OnTextEvent, wx.EVT_TEXT, self)
        PNumEditBoxBase.__init__(self, manager, parent, value,
                                 geometry=geometry, expand=expand)
    
    def set_value(self, value):
        super(PNumEditBox, self).set_value(value)
        self.val.value = value


class PEditControl(PWxEditBase, PEditControlBase):
    
    _style = wx.TE_MULTILINE | wx.TE_PROCESS_TAB
    
    _align = True  # used by panel to determine placement
    _expand = True
    
    def __init__(self, manager, parent, text="",
                 scrolling=False, font=None):
        
        if scrolling:
            self._style = self._style | ec_scroll_style
        PWxEditBase.__init__(self, parent, True)  # automatically expands
        PEditControlBase.__init__(self, manager, parent, text,
                                  scrolling=scrolling, font=font)
        
        self._overwrite_mode = False  # since wx does not have a built-in property for this
        self._modified = False  # since wx does not have a built-in text mod change event
        self._selection = False  # since wx does not have a built-in text state change event
        self._mpending = False  # since the EVT_CHAR handler gets called before the modification state is updated
        self._spending = False  # since the EVT_CHAR handler gets called before the selection changes
        # Bypass the normal setup_notify mechanism because we don't want a wrapper
        eventManager.Register(self.OnKeyEvent, wx.EVT_CHAR, self)
        eventManager.Register(self.OnIdleEvent, wx.EVT_IDLE, self)
    
    def OnKeyEvent(self, event):
        # Can't actually test for mod/selection change yet since the event has to be handled first
        self._mpending = True
        self._spending = True
        event.Skip()
    
    def OnIdleEvent(self, event):
        if self._mpending:
            # A key was pressed so check the modification state
            modified = self.IsModified()
            if modified != self._modified:
                self._modified = modified
                self.forward_event(SIGNAL_TEXTMODCHANGED, event, None)
            self._mpending = False
        # Wait until next idle event if both pending flags were set
        elif self._spending:
            # The super call above changed the selection state, so now we can test it
            selection = bool(self.GetStringSelection())
            if selection != self._selection:
                self._selection = selection
                self.forward_event(SIGNAL_TEXTSTATECHANGED, event, None)
            self._spending = False
        event.Skip()  # FIXME: this doesn't seem to be needed?
    
    # FIXME: no way to programmatically get or set overwrite mode in wx? Ins key works
    
    def GetOverwriteMode(self):
        return self._overwrite_mode
    
    def SetOverwriteMode(self, value):
        if value != self._overwrite_mode:
            # TODO: Fake an Ins key press
            pass
        self._overwrite_mode = value
    
    def setup_scrolling(self, scrolling):
        pass  # this must be done in the constructor in wxWidgets
    
    fn_get_overwrite = 'GetOverwriteMode'
    fn_set_overwrite = 'SetOverwriteMode'
    
    # FIXME: No undo facility in wx on non-Windows platforms
    
    def can_undo(self):
        return self.CanUndo()
    
    def can_redo(self):
        return self.CanRedo()
    
    def can_clip(self):
        return self.CanCut() or self.CanCopy()
    
    def can_paste(self):
        return self.CanPaste()
    
    def clear_edit(self):
        self.Clear()
    
    def undo_last(self):
        self.Undo()
    
    def redo_last(self):
        self.Redo()
    
    def select_all(self):
        self.SelectAll()
    
    def clear_selection(self):
        self.SelectNone()
    
    def delete_last_cut(self):
        pass  # wx handles this differently, see next method
    
    def delete_selected(self):
        self.Remove(*self.GetSelection())
    
    def copy_to_clipboard(self):
        self.Cut()
    
    def cut_to_clipboard(self):
        self.Copy()
    
    def paste_from_clipboard(self):
        self.Paste()



#!/usr/bin/env python3
"""
Module TABWIDGET -- UI Tab Widget
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.stdlib.coll import baselist

from plib.ui.defs import *
from plib.ui.widgets import widget_from_spec

from .app import PWidgetBase


class PTabWidgetBase(PWidgetBase, baselist):
    """Tab widget as a Python list of 2-tuples (tab-title, widget).
    """
    
    signals = (
        SIGNAL_TABSELECTED,
    )
    
    def __init__(self, manager, parent, tabs=None):
        PWidgetBase.__init__(self, manager, parent)
        baselist.__init__(self)
        self.create_tabs(tabs)
    
    def create_tabs(self, tabs):
        """Create tabs from sequence.
        
        Assumes that the sequence contains tuples of (title, spec)
        for each tab.
        """
        self.extend((title, widget_from_spec(self.manager, self, *spec)) for title, spec in tabs)
    
    def append_and_focus(self, tab):
        """Add new tab and set focus to it.
        
        Assumes that tab is a (title, widget) tuple.
        """
        self.append(tab)
        self.set_current_index(len(self) - 1)
    
    def __len__(self):
        return self.tab_count()
    
    def _indexlen(self):
        return self.tab_count()
    
    def tab_count(self):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def get_tab_title(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def set_tab_title(self, index, title):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def tab_at(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def add_tab(self, index, title, widget):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def del_tab(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _get_data(self, index):
        return (self.get_tab_title(index), self.tab_at(index))
    
    def _set_data(self, index, value):
        # This gets a little complicated because we have to delete the
        # old tab at this index and insert a new one.
        self._del_data(index)
        self._add_data(index, value)
    
    def _add_data(self, index, value):
        self.add_tab(index, value[0], value[1])
    
    def _del_data(self, index):
        self.del_tab(index)
    
    def current_index(self):
        """Return the currently selected tab index.
        
        Derived classes must implement.
        """
        raise NotImplementedError
    
    def set_current_index(self, index):
        """Set the currently selected tab index.
        
        Derived classes must implement.
        """
        raise NotImplementedError
    
    def current_title(self):
        return self[self.current_index()][0]
    
    def current_widget(self):
        return self[self.current_index()][1]



#!/usr/bin/env python3
"""
Module QT5.TABWIDGET -- Python Qt 5 Tab Widget
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for the tab widget.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.tabwidget import PTabWidgetBase

from .app import PQtSequenceMeta, PQtWidgetBase


class PTabWidget(qt.QTabWidget, PQtWidgetBase, PTabWidgetBase,
                 metaclass=PQtSequenceMeta):
    
    widget_class = qt.QTabWidget
    
    def __init__(self, manager, parent, tabs=None):
        self._item = None
        self._target = None
        self._setting_index = False
        qt.QTabWidget.__init__(self, parent)
        PTabWidgetBase.__init__(self, manager, parent, tabs=tabs)
    
    def count(self, value):
        # Method name collision, we want it to be the Python sequence
        # count method here.
        return PTabWidgetBase.count(self, value)
    
    def tab_count(self):
        # Let this method access the Qt tab widget count method.
        return qt.QTabWidget.count(self)
    
    def get_tab_title(self, index):
        return str(self.tabText(index))
    
    def set_tab_title(self, index, title):
        self.setTabText(index, str(title))
    
    def tab_at(self, index):
        return self.widget(index)
    
    def add_tab(self, index, title, widget):
        self.insertTab(index, widget, str(title))
    
    def del_tab(self, index):
        self.removeTab(index)
    
    def current_index(self):
        return self.currentIndex()
    
    def _current_changed(self, index):
        # Wrapper for tab changed signal.
        if (not self._setting_index) and self._target:
            self._target(index)
    
    def connect_target(self, signal, target):
        super_method = super(PTabWidget, self).connect_target
        if signal == SIGNAL_TABSELECTED:
            # Hack to capture double firing of tab changed signal when
            # tab is changed programmatically instead of by user
            self._target = target
            super_method(signal, self._current_changed)
        else:
            super_method(signal, target)
    
    def set_current_index(self, index):
        # Wrap the call to avoid double calling of signal handler, then
        # make the call by hand.
        self._setting_index = True
        self.setCurrentIndex(index)
        self._setting_index = False
        self._current_changed(index)



#!/usr/bin/env python3
"""
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This sub-package contains the Qt 5 toolkit modules.
"""



#!/usr/bin/env python3
"""
Module WX.TABLE -- Python wxWidgets Table Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for table table.
"""

import functools

import wx.grid

from plib.ui.defs import *
from plib.ui.base.table import PTableLabelsBase, PTableBase, scrollbarsize, gridlinesize

from .app import PWxSequenceWidget, align_map


class PWxTableLabels(PTableLabelsBase):
    
    _contents = None
    _stretch = None
    
    def _set_label(self, index, label):
        self.table.SetColLabelValue(index, label)
    
    def _set_width(self, index, width):
        if width == WIDTH_STRETCH:
            # Can't compute stretch until all column widths are set
            if self._stretch is None:
                self._stretch = set()
            self._stretch.add(index)
        elif width == WIDTH_CONTENTS:
            # Can't compute sizing to contents until data is added
            if self._contents is None:
                self._contents = set()
            self._contents.add(index)
        elif width > 0:
            self.table.SetColSize(index, width)
    
    def _set_align(self, index, align):
        self.table.SetColLabelAlignment(align_map[align], wx.ALIGN_CENTER_VERTICAL)
    
    def _set_readonly(self, index, readonly):
        pass


class PTable(PWxSequenceWidget, wx.grid.Grid, PTableBase):
    
    labels_class = PWxTableLabels
    
    _align = True  # used by panel to determine placement
    _expand = True
    
    _initialized = False
    
    def __init__(self, manager, parent, labels=None, data=None,
                 font=None, header_font=None):
        
        wx.grid.Grid.__init__(self, parent)
        self.CreateGrid(0, 0)
        PTableBase.__init__(self, manager, parent, labels=labels, data=data,
                            font=font, header_font=header_font)
        
        self.SetColLabelSize(wx.grid.GRID_AUTOSIZE)
        self.SetRowLabelSize(wx.grid.GRID_AUTOSIZE)
        self.SizeColumns()
        self._initialized = True  # so adding/deleting data now triggers SizeColumns
        
        # FIXME: We have to go through all this rigmarole because, first,
        # wx doesn't track the previous row/col in the cell selected event,
        # and second, wx fires the cell selected event before it updates the
        # widget methods that give the current row/col, so we have to roll
        # our own to allow event handlers to see the correct state, and third,
        # wx fires the cell selected event *before* the cell changed event if
        # you're editing a cell and then navigate to a new one
        self._editing_cell = False
        self._pending_select = None
        self.setup_notify(SIGNAL_TABLECELLSELECTED, self.OnCellSelected)
        self.setup_notify(SIGNAL_TABLECELLEDITING, self.OnCellEdit)
        self.setup_notify(SIGNAL_TABLECELLEDITDONE, self.OnCellEditDone)
        self.setup_notify(SIGNAL_TABLECHANGED, self.OnCellChanged)
        if self.labels._stretch:
            self.setup_notify(SIGNAL_SIZEEVENT, self.OnSizeEvent)
    
    def SizeColumns(self):
        # This handles the "fit to contents" columns; it has to be done first so the
        # stretch logic is using the correct values
        contents_indexes = self.labels._contents
        if not contents_indexes:
            return
        for col in contents_indexes:
            self.AutoSizeColumn(col, True)
        
        # This is factored out because it needs to be redone on size event
        self.StretchColumns()
    
    def _add(self, index, value):
        super(PTable, self)._add(index, value)
        if self._initialized:
            self.SizeColumns()
    
    def _del(self, index):
        super(PTable, self)._del(index, value)
        if self._initialized:
            self.SizeColumns()
    
    def OnCellSelected(self, event):
        new_row = event.GetRow()
        new_col = event.GetCol()
        if self._editing_cell:
            self._pending_select = (new_row, new_col)
        else:
            selected_fields = (
                ('CurrentRow', new_row),
                ('CurrentCol', new_col),
                ('PreviousRow', self.current_row()),
                ('PreviousCol', self.current_col()),
            )
            self.forward_event(SIGNAL_CELLSELECTED, event, selected_fields)
    
    def OnCellEdit(self, event):
        self._editing_cell = True
    
    # Temporary vars used by ugly hack in wrap_target below
    _row = None
    _col = None
    
    def OnCellEditDone(self, event):
        assert self._editing_cell
        self._row = event.GetRow()
        self._col = event.GetCol()
        changed_fields = (
            ('Row', self._row),
            ('Col', self._col),
        )
        self.forward_event(SIGNAL_CELLCHANGED, event, changed_fields)
    
    def wrap_target(self, signal, target):
        # Ugly hack so current_row and current_col return the correct values
        # inside the SIGNAL_CELLCHANGED handler, even if a new cell was
        # selected so the cell selected event fired first and updated the
        # wx grid's idea of what the current cell is
        target = super(PTable, self).wrap_target(signal, target)
        if signal == SIGNAL_CELLCHANGED:
            @functools.wraps(target)
            def _hack(*args, **kwargs):
                result = target(*args, **kwargs)
                self._row = None
                self._col = None
                return result
            return _hack
        return target
    
    def OnCellChanged(self, event):
        assert self._editing_cell
        self._editing_cell = False
        if self._pending_select:
            new_row, new_col = self._pending_select
            self._pending_select = None
            selected_fields = (
                ('CurrentRow', new_row),
                ('CurrentCol', new_col),
                # We cheat here by using the cell changed row/col as the previous row/col
                # for the cell select event (it would be really nice if wx would do this for us)
                ('PreviousRow', event.GetRow()),
                ('PreviousCol', event.GetCol()),
            )
            self.forward_event(SIGNAL_CELLSELECTED, event, selected_fields)
    
    def OnSizeEvent(self, event):
        self.StretchColumns()
    
    def StretchColumns(self):
        stretch_indexes = self.labels._stretch
        if not stretch_indexes:
            return
        for col in stretch_indexes:
            self.AutoSizeColumn(col, True)
        if self.get_width() > self.minwidth():
            stretch_count = len(stretch_indexes)
            base_width = sum(self.colwidth(col) for col in range(self.colcount()) if col not in stretch_indexes)
            border_width = self.leftmargin() + scrollbarsize + ((self.colcount() - stretch_count) * gridlinesize)
            stretch_width = (self.get_width() - base_width - border_width) // stretch_count
            for col in stretch_indexes:
                if stretch_width > self.colwidth(col):
                    self.labels._set_width(col, stretch_width)
    
    def set_header_font_object(self, font_name, font_size, bold, italic):
        self.SetLabelFont(self.wx_font_object(
            font_name, font_size, bold, italic
        ))
    
    def _get_cell(self, row, col):
        return self.GetCellValue(row, col)
    
    def _set_cell(self, row, col, value):
        self.SetCellValue(row, col, str(value))
    
    def rowcount(self):
        return self.GetNumberRows()
    
    def colcount(self):
        return self.GetNumberCols()
    
    def set_colcount(self, count):
        self.InsertCols(0, count)
    
    def current_row(self):
        # Must check above ugly hack for SIGNAL_CELLCHANGED first
        return self._row or self.GetGridCursorRow()
    
    def current_col(self):
        # Must check above ugly hack for SIGNAL_CELLCHANGED first
        return self._col or self.GetGridCursorCol()
    
    def _insert_row(self, index):
        self.InsertRows(index, 1)
    
    def _remove_row(self, index):
        self.DeleteRows(index, 1)
    
    def topmargin(self):
        return self.GetColLabelSize()
    
    def leftmargin(self):
        return self.GetRowLabelSize()
    
    def rowheight(self, row):
        return self.GetRowSize(row)
    
    def colwidth(self, col):
        return self.GetColSize(col)
    
    def force_repaint(self):
        self.Refresh()
    
    def default_fgcolor(self):
        return self.GetDefaultCellTextColour()
    
    def default_bkcolor(self):
        return self.GetDefaultCellBackgroundColour()
    
    def set_text_fgcolor(self, row, col, color):
        self.SetCellTextColour(row, col, self._mapped_color(color))
    
    def set_cell_bkcolor(self, row, col, color):
        self.SetCellBackgroundColour(row, col, self._mapped_color(color))



#!/usr/bin/env python3
"""
Module DIALOGS -- UI Dialog Runner and Standard Dialogs
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.widgets import *


class DialogRunner:
    
    dialog_class = get_toolkit_class('dialog', "PDialog")
    
    dialog_client = None
    
    def __init__(self, app, caption, callback=None):
        self.app = app
        self.caption = caption
        self.callback = callback
    
    def run(self):
        dialog = self.dialog_class(self, self.app.main_window, self.caption, self.dialog_client)
        self.populate()
        dialog.setup_notify(SIGNAL_FINISHED, self.dialog_done)
        dialog.display()
    
    show_dialog = run  # syntactic sugar
    
    def dialog_done(self, accepted):
        if accepted and self.callback:
            self.callback(self.get_result())
    
    def populate(self):
        pass
    
    def get_result(self):
        return None


class DisplayDialog(DialogRunner):
    
    def __init__(self, app, caption, client=None):
        DialogRunner.__init__(self, app, caption)
        if client:
            self.dialog_client = client


class StringSelectDialog(DialogRunner):
    
    dialog_client = frame(ALIGN_JUST, LAYOUT_VERTICAL, [
        panel(ALIGN_JUST, LAYOUT_VERTICAL, [
            listbox('values', []),
        ]),
        panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
            action_button(ACTION_OK),
            action_button(ACTION_CANCEL),
        ]),
    ])
    
    def __init__(self, app, caption, values, starting_value=None, callback=None):
        DialogRunner.__init__(self, app, caption, callback)
        self.values = values
        self.starting_value = starting_value
    
    def populate(self):
        self.listbox_values.extend(self.values)
        if self.starting_value:
            self.listbox_values.set_current_text(self.starting_value)
    
    def get_result(self):
        return self.listbox_values.current_text()



#!/usr/bin/env python3
"""
Module QT5.DISPLAY -- Python Qt 5 Text Display Widgets
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for text display widgets.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.base.display import PTextDisplayBase

from .app import PQtWidgetMeta, PQtWidgetBase


class PTextDisplay(qt.QPlainTextEdit, PQtWidgetBase, PTextDisplayBase,
                   metaclass=PQtWidgetMeta):
    
    fix_width_on_resize = True
    fix_height_on_resize = True
    
    widget_class = qt.QPlainTextEdit
    
    def __init__(self, manager, parent, text=None,
                 scrolling=False, font=None):
        
        qt.QPlainTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        PTextDisplayBase.__init__(self, manager, parent, text,
                                  scrolling=scrolling, font=font)
    
    def setup_scrolling(self, scrolling):
        if scrolling:
            self.setLineWrapMode(qt.QPlainTextEdit.NoWrap)
    
    def get_text(self):
        return self.toPlainText()
    
    def set_text(self, value):
        self.setPlainText(value)
    
    def add_text(self, value):
        cursor = self.textCursor()
        cursor.movePosition(cursor.End)
        cursor.insertText(value)
    
    def clear_text(self):
        self.clear()



#!/usr/bin/env python3
"""
Module PYSIDE2.BUTTON -- Python PySide 2 Button Widgets
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for button widgets.
"""

from PySide2 import QtGui as qtg, QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.button import PButtonBase, PActionButtonBase, PCheckBoxBase

from .app import PQtActionMixin, PQtWidget


class PQtButtonBase(PQtActionMixin, PQtWidget, qt.QPushButton):
    
    fix_width_on_resize = True
    
    def __init__(self, parent,
                 geometry=None):
        
        qt.QPushButton.__init__(self, parent)
        self.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
    
    def set_caption(self, caption):
        self.setText(caption)
    
    def set_icon_obj(self, icon):
        self.setIcon(icon)


class PButton(PQtButtonBase, PButtonBase):
    
    def __init__(self, manager, parent, caption, icon=None,
                 geometry=None):
        
        PQtButtonBase.__init__(self, parent)
        PButtonBase.__init__(self, manager, parent, caption, icon,
                             geometry=geometry)


class PActionButton(PQtButtonBase, PActionButtonBase):
    
    event_signals = (SIGNAL_ACTIVATED,)
    
    def __init__(self, manager, parent, action,
                 geometry=None):
        
        PQtButtonBase.__init__(self, parent)
        PActionButtonBase.__init__(self, manager, parent, action,
                                   geometry=geometry)


class PCheckBox(PQtWidget, qt.QCheckBox, PCheckBoxBase):
    
    fn_get_checked = 'isChecked'
    fn_set_checked = 'setChecked'
    
    def __init__(self, manager, parent, label, checked=None,
                 geometry=None, tristate=False):
        
        qt.QCheckBox.__init__(self, parent)
        self.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
        PCheckBoxBase.__init__(self, manager, parent, label, checked=checked,
                               geometry=geometry, tristate=tristate)
    
    def set_caption(self, caption):
        self.setText(caption)
    
    def make_tristate(self):
        self.setTriState(True)



#!/usr/bin/env python3
"""
Module EDITCTRL -- UI Editing Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import functools

from plib.ui.defs import *

from .app import PWidgetBase


class PEditBase(PWidgetBase):
    """Base edit control class, defines standard API.
    """
    
    fn_get_text = None
    fn_set_text = None
    
    fn_get_read_only = None
    fn_set_read_only = None
    
    def get_text(self):
        return getattr(self, self.fn_get_text)()
    
    def set_text(self, value):
        getattr(self, self.fn_set_text)(value)
    
    edit_text = property(get_text, set_text)
    
    def get_read_only(self):
        return getattr(self, self.fn_get_read_only)()
    
    def set_read_only(self, value):
        getattr(self, self.fn_set_read_only)(value)
    
    read_only = property(get_read_only, set_read_only)


class PEditBoxBase(PEditBase):
    """Base class for single-line input control.
    """
    
    signals = (
        SIGNAL_EDITCHANGED,
        SIGNAL_ENTER,
    )
    
    def __init__(self, manager, parent, text="",
                 geometry=None, expand=True):
        
        self.setup_expand(expand)  # needs to be done before geometry is set
        PEditBase.__init__(self, manager, parent,
                           geometry=geometry)
        self.edit_text = text
    
    def setup_expand(self, expand):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PNumEditBoxBase(PEditBoxBase):
    """Base class for edit box for editing numbers.
    """
    
    def __init__(self, manager, parent, value=0,
                 geometry=None, expand=True):
        
        PEditBoxBase.__init__(self, manager, parent,
                              geometry=geometry, expand=expand)
        self.set_value(value)
    
    def get_value(self):
        return int(self.get_text())
    
    def set_value(self, value):
        if not isinstance(value, int):
            raise TypeError("Numeric edit can only accept int values: {} is not an int".format(repr(value)))
        self.set_text(str(value))
    
    edit_value = property(get_value, set_value)
    
    def wrap_target(self, signal, target):
        if signal == SIGNAL_EDITCHANGED:
            @functools.wraps(target)
            def _wrapper(text):
                target(int(text))
            return _wrapper
        return target


class PEditControlBase(PEditBase):
    """Base class for multi-line edit control.
    """
    
    signals = (
        SIGNAL_TEXTCHANGED,
        SIGNAL_TEXTMODCHANGED,
        SIGNAL_TEXTSTATECHANGED,
    )
    
    def __init__(self, manager, parent, text="",
                 scrolling=False, font=None):
        
        PEditBase.__init__(self, manager, parent,
                           font=font)
        self.setup_scrolling(scrolling)
        self.edit_text = text
    
    fn_get_overwrite = None
    fn_set_overwrite = None
    
    def get_overwrite_mode(self):
        return getattr(self, self.fn_get_overwrite)()
    
    def set_overwrite_mode(self, value):
        getattr(self, self.fn_set_overwrite)(value)
    
    overwrite_mode = property(get_overwrite_mode, set_overwrite_mode)
    
    def setup_scrolling(self, scrolling):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def can_undo(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def can_redo(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def can_clip(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def can_paste(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def clear_edit(self):
        # Derived classes may implement a faster method
        self.edit_text = ""
    
    def undo_last(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def redo_last(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def select_all(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def delete_selected(self):
        self.cut_to_clipboard()
        self.delete_last_cut()
    
    def clear_selection(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def copy_to_clipboard(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def cut_to_clipboard(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def delete_last_cut(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def paste_from_clipboard(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError



#!/usr/bin/env python3
"""
PNOTEPAD.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A simple plain text editor.
"""

import sys
import os

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication
from plib.ui.textfile import PTextFile
from plib.ui.widgets import *


class PMemoFile(PTextFile):
    
    def __init__(self, app, control):
        PTextFile.__init__(self, app, control, [
            ("Text Files", ["txt"]),
            ("All Files", ["*"]),
        ])
        self.main_window = app.main_window
        self.base_title = app.main_title
    
    def set_filename(self, filename):
        super(PMemoFile, self).set_filename(filename)
        if filename:
            self.main_window.set_caption("{} - {}".format(
                self.base_title,
                "<Untitled>" if filename is self.untitled else filename
            ))
        else:
            self.main_window.set_caption(self.base_title)


class PNotepad(PApplication):
    
    about_data = {
        'name': "PNotepad",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "Plain Text Editor",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net",
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    menu_actions = [
        (MENU_FILE, (
            ACTION_FILE_NEW, ACTION_FILE_OPEN, ACTION_FILE_SAVE, ACTION_FILE_SAVEAS, ACTION_FILE_CLOSE, ACTION_EXIT,
        )),
        (MENU_EDIT, (
            ACTION_EDIT_UNDO, ACTION_EDIT_REDO,
            ACTION_EDIT_CUT, ACTION_EDIT_COPY, ACTION_EDIT_PASTE,
            ACTION_EDIT_DELETE, ACTION_EDIT_SELECTALL, ACTION_EDIT_SELECTNONE, ACTION_EDIT_CLEAR,
            ACTION_EDIT_OVERWRITE,
        )),
        (MENU_HELP, (
            ACTION_ABOUT, ACTION_ABOUT_TOOLKIT,
        )),
    ]
    
    toolbar_actions = [
        (ACTION_FILE_NEW, ACTION_FILE_OPEN, ACTION_FILE_SAVE, ACTION_FILE_SAVEAS,),
        (ACTION_EDIT_UNDO, ACTION_EDIT_REDO,),
        (ACTION_EDIT_CUT, ACTION_EDIT_COPY, ACTION_EDIT_PASTE,),
        (ACTION_EDIT_DELETE, ACTION_EDIT_SELECTALL, ACTION_EDIT_SELECTNONE, ACTION_EDIT_CLEAR,),
        (ACTION_EDIT_OVERWRITE,),
        (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT, ACTION_EXIT,),
    ]
    
    status_labels = [
        ('overwrite', "", PANEL_SUNKEN),
        ('size', "0 bytes", PANEL_SUNKEN),
    ]
    
    main_title = "Plain Text Editor"
    main_iconfile = os.path.join(os.path.split(os.path.realpath(__file__))[0],
                                 "pnotepad.png")
    
    main_widget = memo('notepad', font=("Courier New", 12))
    
    def get_enable_list(self):
        actions = self.main_window.actions
        
        enable_list = [
            (getattr(self.memo_notepad, attr), actions[key].enable)
            for attr, key in (
                ('can_undo', ACTION_EDIT_UNDO),
                ('can_redo', ACTION_EDIT_REDO),
                ('can_clip', ACTION_EDIT_CUT),
                ('can_clip', ACTION_EDIT_COPY),
                ('can_paste', ACTION_EDIT_PASTE),
                ('can_clip', ACTION_EDIT_DELETE),
                ('can_clip', ACTION_EDIT_SELECTNONE),
            )
        ]
        
        dirty_list = [
            actions[key].enable
            for key in (
                ACTION_FILE_SAVE,
            )
        ]
        
        return enable_list, dirty_list
    
    def enable_actions(self, changed=None):
        for fstate, fenable in self.enable_list:
            fenable(fstate())
        # We have to check the changed parameter because the signal handler
        # for SIGNAL_TEXTCHANGED might be called before the one in the
        # PTextFile object that updates its dirty state
        dirty = changed if changed is not None else self.memo_file.dirty
        for fenable in self.dirty_list:
            fenable(dirty)
    
    filename_to_open = None
    file_not_found = False
    
    def check_file_open(self):
        if self.filename_to_open:
            filename = self.filename_to_open
            if os.path.isfile(filename):
                print("Opening", filename)
                self.memo_file.open_filename(filename)
            else:
                print(filename, "not found!")
                self.file_not_found = True
    
    def after_create(self):
        self.memo_file = PMemoFile(self, self.memo_notepad)
        self.enable_list, self.dirty_list = self.get_enable_list()
        self.enable_actions()
        self.update_filesize_status()
        self.update_overwrite_status()
        self.check_file_open()
    
    def set_status_text(self, text):
        print(text)
        self.main_window.statusbar.set_text(text)
    
    def update_filesize_status(self,
                               fmt="{} bytes".format):
        
        self.label_size.caption = fmt(self.memo_file.filesize)
    
    overwrite_labels = ("INS", "OVR")
    
    def update_overwrite_status(self):
        self.label_overwrite.caption = self.overwrite_labels[self.memo_notepad.overwrite_mode]
    
    def on_app_shown(self):
        if self.file_not_found:
            self.message_box.error("File Error", "File {} not found".format(self.filename_to_open))
    
    def on_notepad_changed(self):
        self.update_filesize_status()
    
    def on_notepad_mod_changed(self, changed):
        self.enable_actions(changed)
    
    def on_notepad_state_changed(self):
        self.enable_actions()
    
    def check_save(self):
        answer = self.message_box.query3(
            "Unsaved File",
            "Save file?"
        )
        return (
            True if answer == ANSWER_YES else
            False if answer == ANSWER_NO else
            None
        )
    
    def on_file_new(self):
        if self.memo_file.dirty:
            do_save = self.check_save()
            if do_save:
                self.memo_file.save_file()
            if do_save is None:
                return
        self.memo_file.new_file()
        self.enable_actions()
        self.set_status_text("New file")
        self.update_filesize_status()
    
    def on_file_open(self):
        if self.memo_file.open_file():
            self.enable_actions()
            self.set_status_text("Opened file: {}".format(self.memo_file.filename))
            self.update_filesize_status()
    
    def on_file_save(self):
        if self.memo_file.save_file():
            self.enable_actions()
            self.set_status_text("Saved file: {}".format(self.memo_file.filename))
    
    def on_file_saveas(self):
        if self.memo_file.save_file_as():
            self.enable_actions()
            self.set_status_text("Saved file: {}".format(self.memo_file.filename))
    
    def on_file_close(self):
        self.memo_file.close_file()
        self.update_filesize_status()
    
    def on_edit_undo(self):
        self.memo_notepad.undo_last()
        self.enable_actions()
    
    def on_edit_redo(self):
        self.memo_notepad.redo_last()
        self.enable_actions()
    
    def on_edit_cut(self):
        self.memo_notepad.cut_to_clipboard()
        self.enable_actions()
    
    def on_edit_copy(self):
        self.memo_notepad.copy_to_clipboard()
        self.enable_actions()
    
    def on_edit_paste(self):
        self.memo_notepad.paste_from_clipboard()
        self.enable_actions()
    
    def on_edit_delete(self):
        self.memo_notepad.delete_selected()
        self.enable_actions()
    
    def on_edit_selectall(self):
        self.memo_notepad.select_all()
        self.enable_actions()
    
    def on_edit_selectnone(self):
        self.memo_notepad.clear_selection()
        self.enable_actions()
    
    def on_edit_clear(self):
        self.memo_notepad.clear_edit()
        self.enable_actions()
    
    def on_edit_overwrite(self):
        # TODO: wx doesn't seem to support this? (the Ins key works, but no way to detect or control programmatically)
        self.memo_notepad.overwrite_mode = not self.memo_notepad.overwrite_mode
        self.update_overwrite_status()
    
    def on_about(self):
        self.about()
    
    def on_about_toolkit(self):
        self.about_toolkit()
    
    def on_exit(self):
        self.exit_app()
    
    def accept_close(self):
        if self.memo_file.dirty:
            do_save = self.check_save()
            if do_save:
                self.memo_file.save_file()
                return True
            if do_save is None:
                return False
            return self.message_box.query2(
                "Application Exit",
                "Exit anyway?"
            ) == ANSWER_OK
        return True


if __name__ == "__main__":
    if len(sys.argv) > 1:
        PNotepad.filename_to_open = sys.argv[1]
    PNotepad().run()



#!/usr/bin/env python3
"""
Module APP -- UI Application Classes
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.widgets import get_toolkit_class


PApplication = get_toolkit_class('app', 'PApplication')



#!/usr/bin/env python3
"""
PYIDSERVER-UI.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A UI wrapper around the pyidserver example program from
the PLIB.STDLIB package.
"""

import sys
import os
import datetime
import urllib.parse

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication
from plib.ui.output import PTextOutput
from plib.ui.socket import PSocketTransport
from plib.ui.widgets import *


def _load_example_module():
    from plib.stdlib.imp import import_from_path
    from plib.stdlib.postinstall import get_share_dir
    
    return import_from_path(
        os.path.join(
            get_share_dir('plib3.stdlib', 'plib.stdlib'),
            'examples', 'pyidserver'
        ),
        'pyidserver3'
    )


pyidserver = _load_example_module()


PROTOCOL_MAP = {proto: items[0] for proto, items in pyidserver.protocols.items()}

DEFAULT_DNSONLY = pyidserver.run_main.__defaults__[2]
DEFAULT_PROTOCOL = pyidserver.run_main.__defaults__[3]
DEFAULT_PORTNUM = pyidserver.run_main.__defaults__[4]


class IDServerUIClient(pyidserver.IDServerClient):
    
    def __init__(self, transport, fileobj, frame):
        super(IDServerUIClient, self).__init__(transport, fileobj)
        self.frame = frame
    
    def on_close(self):
        super(IDServerUIClient, self).on_close()
        self.frame.done()


CAPTION_GO = "&Go"
ICON_GO = ACTION_OK
CAPTION_STOP = "&Stop"
ICON_STOP = ACTION_CANCEL


class IDServerApp(PApplication):
    
    about_data = {
        'name': "PyIDServer",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "A Python UI for IDServer",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net",
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    main_title = "PyIDServer"
    main_iconfile = os.path.join(os.path.split(os.path.realpath(__file__))[0],
                                 "pyidserver.png")
    
    main_size = SIZE_OFFSET
    width_offset = 640
    main_placement = MOVE_CENTER
    
    main_widget = frame(ALIGN_JUST, LAYOUT_VERTICAL, [
        panel(ALIGN_TOP, LAYOUT_HORIZONTAL, [
            edit('url'),
            button('go', CAPTION_GO, ICON_GO)
        ]),
        panel(ALIGN_JUST, LAYOUT_VERTICAL, [
            box(ALIGN_TOP, LAYOUT_HORIZONTAL, [
                panel(ALIGN_LEFT, LAYOUT_HORIZONTAL, [
                    checkbox('dnsonly', "DNS Only", DEFAULT_DNSONLY),
                    checkbox('protocol', "Set Protocol"),
                    sorted_combo('protocol', pyidserver.protocols.keys(), pyidserver.PROTO_DEFAULT),
                    checkbox('portnum', "Set Port"),
                    num_edit('portnum', DEFAULT_PORTNUM, expand=False)
                ]),
                padding()
            ]),
            text('output', font=("Courier New", 12))
        ]),
        panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
            labeled("Transcript Dir:", label('directory', ""), LAYOUT_HORIZONTAL),
            button('select', "Select Dir"),
            button('save', "Save Transcript"),
            action_button(ACTION_ABOUT),
            action_button(ACTION_ABOUT_TOOLKIT),
            action_button(ACTION_EXIT)
        ])
    ])
    
    protocol_map = PROTOCOL_MAP
    
    default_protocol = DEFAULT_PROTOCOL
    default_portnum = DEFAULT_PORTNUM
    
    create_done = False
    client = None
    
    @property
    def outputs_on(self):
        return self.create_done and not self.client
    
    def after_create(self):
        """This method is called after all widgets are created, but before the app is shown.
        """
        
        # Controls affected by the DNS Only checkbox
        self.dnsonly_controls = (
            (self.checkbox_protocol, self.combo_protocol),
            (self.checkbox_portnum, self.edit_portnum)
        )
        
        # Controls independent of DNS Only checkbox
        self.other_controls = (
            self.edit_url,
            self.checkbox_dnsonly
        )
        
        self.on_dnsonly_toggled()  # technically not needed, put in for completeness
        self.on_protocol_toggled()  # initialize enable state of protocol combo
        self.on_portnum_toggled()  # initialize enable state of portnum edit
        
        # Set up socket transport
        self.transport = PSocketTransport()
        
        # Set up output file-like object here for convenience
        self.outputfile = PTextOutput(self.text_output)
        
        # Initialize transcript save dir
        self.transcript_dir = os.getcwd()
        self.update_dir_display()
        self.button_save.enabled = False
        
        # Start with keyboard focus in the URL text entry
        self.edit_url.set_focus()
        
        # Turn on diagnostic outputs
        self.create_done = True
    
    def on_dnsonly_toggled(self, checked=None):
        """Called when the dns_only checkbox is checked/unchecked.
        
        Only enable protocol and port controls if not DNS only.
        """
        
        if checked is not None:
            assert checked is self.checkbox_dnsonly.checked
        enable = not self.checkbox_dnsonly.checked
        for ctrl, subctrl in self.dnsonly_controls:
            ctrl.enabled = enable
            subctrl.enabled = enable and ctrl.checked
        if self.outputs_on:
            print("DNS Only", ("disabled", "enabled")[not enable])
    
    def on_protocol_toggled(self, checked=None):
        """Called when the protocol checkbox is checked/unchecked.
        
        Sync protocol combo enable with check box."""
        
        if checked is not None:
            assert checked is self.checkbox_protocol.checked
        self.combo_protocol.enabled = enable = self.checkbox_protocol.checked
        if self.outputs_on:
            print("Protocol selection", ("disabled", "enabled")[enable])
    
    def on_protocol_selected(self, index):
        """Called when a protocol combo selection is made.
        
        For now, just prints a diagnostic showing the signal response.
        """
        
        protocol = self.combo_protocol[index]
        if self.outputs_on:
            print("Protocol selected:", index, protocol)
        self.edit_portnum.edit_value = self.protocol_map[protocol]
    
    def on_portnum_toggled(self, checked=None):
        """Called when the portnum checkbox is checked/unchecked.
        
        Sync portnum edit enable with check box.
        """
        
        if checked is not None:
            assert checked is self.checkbox_portnum.checked
        self.edit_portnum.enabled = enable = self.checkbox_portnum.checked
        if self.outputs_on:
            print("Port selection", ("disabled", "enabled")[enable])
    
    def on_portnum_changed(self, value):
        """Called when a port number is entered.
        
        For now, just prints a diagnostic showing what was entered.
        """
        
        assert value == self.edit_portnum.edit_value
        if self.outputs_on:
            print("Port number entered:", self.edit_portnum.edit_value)
    
    def on_go_clicked(self):
        """Called when the Go/Stop button is pushed.
        
        Execute the idserver query, or stop a query in progress.
        For executing a query, this method is also called when the
        Enter key is pressed while in the URL edit box.
        """
        
        if self.client:
            print("Query in progress, stopping")
            self.client.close()
        
        else:
            print("Starting query")
            
            # Clear output
            self.outputfile.truncate()
            
            # Check URL
            url = self.edit_url.edit_text
            if len(url) < 1:
                self.outputfile.write("Error: No URL entered.")
                self.outputfile.flush()
                return
            
            # Fill in arguments that user selected, if any
            dns_only = self.checkbox_dnsonly.checked
            if self.checkbox_protocol.checked:
                protocol = self.combo_protocol.current_text()
            else:
                protocol = self.default_protocol
            if self.checkbox_portnum.checked:
                portnum = self.edit_portnum.edit_value
            else:
                portnum = self.default_portnum
            
            # Now execute
            for ctrl, subctrl in self.dnsonly_controls:
                ctrl.enabled = subctrl.enabled = False
            for ctrl in self.other_controls:
                ctrl.enabled = False
            self.button_go.set_caption(CAPTION_STOP)
            self.button_go.set_icon(ICON_STOP)
            
            self.client = IDServerUIClient(self.transport, self.outputfile, self)
            self.client.run(url, dns_only, protocol, portnum)
            
            print("Query done")
    
    def done(self):
        # Either we're done, or we've stopped a query in progress
        self.button_save.enabled = True
        self.button_go.set_caption(CAPTION_GO)
        self.button_go.set_icon(ICON_GO)
        for ctrl in self.other_controls:
            ctrl.enabled = True
        self.on_dnsonly_toggled()
        self.client = None
    
    on_url_enter = on_go_clicked  # so hitting Enter in the URL edit box also starts the query
    
    def update_dir_display(self):
        self.label_directory.caption = self.transcript_dir
    
    def on_select_clicked(self):
        new_dir = self.file_dialog.choose_directory(self.transcript_dir)
        if new_dir:
            self.transcript_dir = new_dir
            self.update_dir_display()
    
    def on_save_clicked(self):
        url = self.edit_url.edit_text
        if len(url) < 1:
            self.message_box.info("No Data", "No URL has been queried, nothing to save")
            return
        data = self.outputfile.read()
        if len(data) < 1:
            self.message_box.info("No Data", "No transcript data, nothing to save")
            return
        desc = url
        if self.checkbox_protocol.checked:
            desc = "{}://{}".format(self.combo_protocol.current_text(), desc)
        if self.checkbox_portnum.checked:
            desc = "{}:{}".format(desc, self.edit_portnum.edit_value)
        desc = urllib.parse.quote(desc, safe="")  # we want to quote any slashes as well
        filename = "pyidserver-ui_{}_{}.txt".format(desc, round(datetime.datetime.now().timestamp()))
        data = self.outputfile.read()
        with open(os.path.join(self.transcript_dir, filename), 'w') as f:
            f.write(data)
    
    def on_about(self):
        self.about()
    
    def on_about_toolkit(self):
        self.about_toolkit()
    
    def on_exit(self):
        self.exit_app()
    
    def on_app_closing(self):
        # If query is in progress, shut it down before closing
        if self.client:
            self.client.close()


if __name__ == "__main__":
    IDServerApp().run()



#!/usr/bin/env python3
"""
Module PYSIDE2.IMAGEVIEW -- Python PySide 2 Image View Objects
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for the panel widgets.
"""

from PySide2 import QtCore as qtc, QtGui as qtg, QtWidgets as qt

from plib.ui.base.imageview import PImageViewBase

from .app import PQtWidget


class PImageView(PQtWidget, qt.QScrollArea, PImageViewBase):
    
    image_view = None
    
    def __init__(self, manager, parent, filename=None, geometry=None):
        qt.QScrollArea.__init__(self, parent)
        self.setAlignment(qtc.Qt.AlignTop | qtc.Qt.AlignLeft)
        PImageViewBase.__init__(self, manager, parent, filename=filename, geometry=geometry)
    
    @staticmethod
    def supported_formats():
        # This must not be called until after the application object is created
        return [bytes(f).decode() for f in qtg.QImageReader.supportedImageFormats()]
    
    def load_from_file(self, filename):
        pixmap = qtg.QPixmap(filename)
        self.image_view = qt.QLabel()
        self.image_view.setAlignment(qtc.Qt.AlignTop | qtc.Qt.AlignLeft)
        self.image_view.setPixmap(pixmap)
        self.setWidget(self.image_view)
    
    def get_image_size(self):
        size = self.image_view.pixmap().size()
        return size.width(), size.height()
    
    def zoom_to(self, width, height):
        pixmap = self.image_view.pixmap()
        self.image_view.setPixmap(pixmap.scaled(width, height, qtc.Qt.KeepAspectRatio))
        self.image_view.resize(width, height)
    
    def rotate_90(self, clockwise):
        width, height = self.get_image_size()
        transform = qtg.QTransform()
        transform.rotate(90.0 * (1.0 if clockwise else -1.0))
        pixmap = self.image_view.pixmap()
        self.image_view.setPixmap(pixmap.transformed(transform))
        self.image_view.resize(height, width)



#!/usr/bin/env python3
"""
Module QT5.LISTVIEW -- Python Qt 5 Tree/List View Objects
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for the tree/list view widgets.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.base.listview import (
    PTreeViewItemBase, PListViewLabelsBase,
    PListViewItemBase, PSortedListViewItemBase,
    PTreeViewBase, PListViewBase, PSortedListViewBase
)

from .app import PQtSequenceMeta, PQtSignalBase, PQtWidgetBase, align_map
from .header import PQtHeaderBase, PQtHeaderWidget


class PQtListViewItemBase(qt.QTreeWidgetItem, PQtSignalBase):
    
    def item_count(self):
        return self.childCount()
    
    def item_at(self, index):
        return self.child(index)
    
    def index_of(self, item):
        return self.indexOfChild(item)
    
    def _add_item(self, index, item):
        if index == len(self):
            self.addChild(item)
        else:
            self.insertChild(index, item)
    
    def _del_item(self, index, item):
        self.takeChild(index)
    
    def _get_col(self, col):
        return str(self.text(col))
    
    def _set_col(self, col, value):
        self.setText(col, str(value))
        # FIXME: it would be nice if this could be done once instead of
        # per item
        self.setTextAlignment(col,
                              self.listview.headerItem().textAlignment(col))


class PTreeViewItem(PQtListViewItemBase, PTreeViewItemBase,
                    metaclass=PQtSequenceMeta):
    
    def __init__(self, parent, index, data=None):
        PQtListViewItemBase.__init__(self)
        PTreeViewItemBase.__init__(self, parent, index, data)
    
    def expand(self):
        self.setExpanded(True)


class PListViewItem(PQtListViewItemBase, PListViewItemBase,
                    metaclass=PQtSequenceMeta):
    
    def __init__(self, parent, index, data=None):
        PQtListViewItemBase.__init__(self)
        PListViewItemBase.__init__(self, parent, index, data)


class PSortedListViewItem(PQtListViewItemBase, PSortedListViewItemBase,
                          metaclass=PQtSequenceMeta):
    
    def __init__(self, parent, index, data=None):
        PQtListViewItemBase.__init__(self)
        PSortedListViewItemBase.__init__(self, parent, index, data)


class PListViewLabels(PQtHeaderBase, PListViewLabelsBase):
    
    def _set_labels_from_list(self, label_list):
        self.listview.setHeaderLabels(label_list)
    
    def _header_item(self, index):
        return self.listview.headerItem()
    
    def _header_view(self):
        return self.listview.header()
    
    def _set_align(self, index, align):
        item = self.listview.headerItem()
        item.setTextAlignment(index, align_map[align])
        # each item will align itself when added


class PQtListViewBase(qt.QTreeWidget, PQtHeaderWidget, PQtWidgetBase):
    
    widget_class = qt.QTreeWidget
    
    labels_class = PListViewLabels
    
    def __init__(self, parent):
        qt.QTreeWidget.__init__(self, parent)
        self.header().setStretchLastSection(False)
        self.setSortingEnabled(False)
        self.setRootIsDecorated(True)
    
    def item_count(self):
        return self.topLevelItemCount()
    
    def item_at(self, index):
        return self.topLevelItem(index)
    
    def index_of(self, item):
        return self.indexOfTopLevelItem(item)
    
    def _add_item(self, index, item):
        if index == len(self):
            self.addTopLevelItem(item)
        else:
            self.insertTopLevelItem(index, item)
    
    def _del_item(self, index, item):
        self.takeTopLevelItem(index)
    
    def header_object(self):
        return self.header()
    
    def colcount(self):
        return self.columnCount()
    
    def current_item(self):
        return self.currentItem()
    
    def set_current_item(self, item):
        self.setCurrentItem(item)


class PTreeView(PQtListViewBase, PTreeViewBase,
                metaclass=PQtSequenceMeta):
    
    item_class = PTreeViewItem
    
    def __init__(self, manager, parent, labels=None, data=None, auto_expand=False,
                 font=None, header_font=None):
        
        PQtListViewBase.__init__(self, parent)
        PTreeViewBase.__init__(self, manager, parent, labels=labels, data=data, auto_expand=auto_expand,
                               font=font, header_font=header_font)


class PListView(PQtListViewBase, PListViewBase,
                metaclass=PQtSequenceMeta):
    
    item_class = PListViewItem
    
    def __init__(self, manager, parent, labels=None, data=None,
                 font=None, header_font=None):
        
        PQtListViewBase.__init__(self, parent)
        PListViewBase.__init__(self, manager, parent, labels=labels, data=data,
                               font=font, header_font=header_font)


class PSortedListView(PQtListViewBase, PSortedListViewBase,
                      metaclass=PQtSequenceMeta):
    
    item_class = PSortedListViewItem
    
    def __init__(self, manager, parent, labels=None, data=None, key=None,
                 font=None, header_font=None):
        
        PQtListViewBase.__init__(self, parent)
        PSortedListViewBase.__init__(self, manager, parent, labels=labels, data=data,
                                     key=key, font=font, header_font=header_font)



#!/usr/bin/env python3
"""
Module HELPERS -- UI Widget Helper Classes
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Defines "helper" classes that implement the basic functionality
to make widgets look like Python containers. The two basic
container structures supported are:

- A "row" structure: the top-level container is a sequence of rows,
  and each row is in turn a sequence of columns. The top-level
  container is fully mutable (new rows can be inserted/appended,
  and existing rows can be deleted), but each row is of fixed length
  (columns cannot be added or deleted).

- A "tree" structure: the top-level container is a mutable list of items,
  where each item is a 2-sequence: the first member is a fixed-length
  sequence of columns, and the second member is a list of sub-items,
  which has the same structure recursively.
"""

from plib.stdlib.coll import abstractlist, basesequence, normalize

from plib.ui.defs import *
from plib.ui.common import font_args


class PHeaderLabel(object):
    """Encapsulates a header label for classes that use the helpers below.
    """
    
    def __init__(self, text, width=WIDTH_CONTENTS, align=ALIGN_LEFT, readonly=False):
        self.text = text
        self.width = width  # positive int, or WIDTH_CONTENTS or WIDTH_STRETCH
        self.align = align
        self.readonly = readonly
    
    def __str__(self):
        return self.text


class _PHelperData(object):
    # Abstract base class for helper items that update displayed
    # text from data.
    
    def __init__(self, data=None):
        if data is not None:
            self._update(data)
    
    def _update(self, data):
        """Derived classes must implement to populate widget from data.
        """
        raise NotImplementedError


class _PHelperItem(_PHelperData):
    # Abstract mixin class for helper items. See the PHelper docstring
    # below for more details.
    
    def __init__(self, index, data=None):
        self._i = index
        _PHelperData.__init__(self, data)
    
    def _value(self):
        """Derived classes must implement to get widget data.
        """
        raise NotImplementedError


class _PColsMixin(basesequence):
    # Mixin class to expose multiple columns of a single row
    # or node in a multi-column widget as a sequence. Designed to
    # be a mixin or helper for a _PHelperItem subclass. Assumes
    # that data is a sequence of strings.
    
    _updating = False
    
    def _update(self, data):
        self._updating = True
        self._len = len(data)
        for index, item in enumerate(data):
            self[index] = item
        self._updating = False
    
    def _indexlen(self):
        return self._len


class _PHelperColLabels(_PColsMixin, _PHelperData):
    # Encapsulates the column labels of a multi-column widget
    # with _PHelper.
    
    def __init__(self, helper, labels=None):
        self._helper = helper
        # Allow each label to be either a string (title only), or a tuple of arguments
        # (to allow width, alignment, and readonly args to be given)
        labels = [PHeaderLabel(*((args,) if isinstance(args, str) else args)) for args in labels]
        _PHelperData.__init__(self, labels)
    
    def _get_data(self, index):
        return self._get_label(index)
    
    def _set_data(self, index, value):
        self._set_label(index, str(value))
        self._set_width(index, value.width)
        self._set_align(index, value.align)
        self._set_readonly(index, value.readonly)
    
    def _get_label(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _set_label(self, index, label):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _set_width(self, index, width):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _set_align(self, index, align):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _set_readonly(self, index, readonly):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class _PHelper(abstractlist):
    """Make a widget look like a list of child items.
    
    Abstract class to provide common methods to make a widget look
    like a list of 'child' items. Each 'child' item should be a mixin
    of a _PHelperItem (see above) and either a child widget (such as a
    PListViewItem child of a PListView) or an object that manages a
    portion of the parent widget (such as a PTableRow for managing a
    single row of a PTable).
    
    Note that the basic method of implementation for this class is to
    override key list methods to provide "hooks" for side effects to
    occur when items are added or deleted. The "hook" methods must then
    be overridden in derived/mixin classes to implement the specific
    side effects desired (such as adding/deleting items from a list
    view or rows from a table).
    
    Note also that most of the sequence emulation methods work with the
    actual 'values', not the helper items. To access the helper items,
    use the items attribute (which is a standard dict mapping index
    values to the helper items).
    """
    
    labels_class = None
    
    def __init__(self, labels=None, data=None, header_font=None):
        if header_font is not None:
            # Do this before adding data
            if isinstance(header_font, str):
                header_font = (header_font,)  # font name only is allowed
            self.set_header_font(*header_font)
        if (labels is not None) and (self.labels_class is not None):
            self.labels = self.labels_class(self, labels)
        else:
            self.labels = []
        abstractlist.__init__(self, data)
    
    def __len__(self):
        return self.item_count()
    
    def _indexlen(self):
        return self.item_count()
    
    def item_count(self):
        """Return the number of items.
        
        Derived classes must override to implement.
        """
        raise NotImplementedError
    
    def item_at(self, index):
        """Return the item at the given index.
        
        Derived classes must override to implement.
        """
        raise NotImplementedError
    
    def index_of(self, item):
        """Return the index of the item.
        
        Derived classes must override to implement.
        """
        raise NotImplementedError
    
    def all_items(self):
        for index in range(self.item_count()):
            yield self.item_at(index)
    
    def set_header_font(self, font_name=None, font_size=None,
                        bold=None, italic=None):
        
        self.set_header_font_object(*font_args(
            self,
            font_name,
            font_size,
            bold,
            italic
        ))
    
    def set_header_font_object(self, font_name, font_size, bold, italic):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _newitem(self, index, value):
        # Derived classes can override to change the signature of
        # the item class or substitute another callable that returns
        # an item instance.
        return self.item_class(self, index, value)
    
    def _get(self, index):
        # Return the value of the helper item at index. Note that we assume
        # each item stored in the items dict is a _PHelperItem or derived from it.
        return self.item_at(index)._value()
    
    def _set(self, index, value):
        # Update helper item at index with new value. Note that this method
        # should never be called with an index out of range (meaning that the
        # index will always be in the items dict); adding a new item should
        # always be done using the _add method.
        self.item_at(index)._update(value)
    
    def _add_item(self, index, item):
        # Add new item at index
        raise NotImplementedError
    
    def _add(self, index, value):
        # Insert or append a helper item wrapped around value.
        item = self._newitem(index, value)
        self._add_item(index, item)
    
    def _del_item(self, index, item):
        # Delete item at index
        raise NotImplementedError
    
    def _del(self, index):
        # Delete the helper item at index.
        item = self.item_at(index)
        self._del_item(index, item)
    
    # Note: we subclass abstractlist instead of baselist because we don't
    # want to support slicing
    
    def __getitem__(self, index):
        if isinstance(index, int):
            return self._get(normalize(self.__len__(), index))
        else:
            raise TypeError("List index must be an int.")
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            self._set(normalize(self.__len__(), index), value)
        else:
            raise TypeError("List index must be an int.")
    
    def __delitem__(self, index):
        if isinstance(index, int):
            self._del(normalize(self.__len__(), index))
        else:
            raise TypeError("List index must be an int.")
    
    def insert(self, index, value):
        if isinstance(index, int):
            self._add(normalize(self.__len__(), index, noraise=True), value)
        else:
            raise TypeError("List index must be an int.")


class _PRowHelperItem(_PColsMixin, _PHelperItem):
    """Helper item that exposes a sequence of columns.
    
    Intended to map to a row in a multi-column widget which does *not*
    have nested tree functionality, such as a table.
    """
    
    def __init__(self, helper, index, data=None):
        self._helper = helper
        _PHelperItem.__init__(self, index, data)
    
    def _value(self):
        # This item already appears as a sequence of column values
        return self


class _PRowHelper(_PHelper):
    """Helper class for list or table-type widgets.
     
    Assumes that each row has a helper item, but no nested tree structure.
    """
    
    def __init__(self, labels=None, data=None, header_font=None):
        self._rowlist = []  # these widgets won't store rows so we have to
        _PHelper.__init__(self, labels, data, header_font=header_font)
    
    def item_count(self):
        result = self.rowcount()
        assert result == len(self._rowlist)
        return result
    
    def index_of(self, item):
        return self._rowlist.index(item)
    
    def item_at(self, index):
        return self._rowlist[index]
    
    def _add(self, index, value):
        self._insert_row(index)
        _PHelper._add(self, index, value)
    
    def _add_item(self, index, item):
        self._rowlist.insert(index, item)
    
    def _del(self, index):
        _PHelper._del(self, index)
        self._remove_row(index)
    
    def _del_item(self, index):
        del self._rowlist[index]
    
    def _insert_row(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _remove_row(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def rowcount(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def current_row(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def current_item(self):
        r = self.current_row()
        if r < 0:
            return None
        else:
            return self.item_at(r)


class _PTreeHelper(_PHelper):
    """Helper class for tree-type widgets.
    
    Provides methods to easily manage a tree with this object as root.
    """
    
    def current_item(self):
        """Return the currently selected helper item.
        
        Derived classes must override to implement.
        """
        raise NotImplementedError
    
    def set_current_item(self, item):
        """Set the currently selected helper item.
        
        Derived classes must override to implement.
        """
        raise NotImplementedError
    
    def current_index(self):
        item = self.current_item()
        if item is not None:
            return self.index_of(item)
        return None
    
    def set_current_index(self, index):
        self.set_current_item(self.item_at(index))


class _PTreeHelperItemCols(_PColsMixin, _PHelperData):
    """Expose columns of a tree view node as a sequence.
    """
    
    def __init__(self, helperitem, data=None):
        self._helperitem = helperitem
        _PHelperData.__init__(self, data)


class _PTreeHelperItem(_PHelperItem, _PTreeHelper):
    """Tree helper item that is also the root of its own subtree.
    
    Helper item that appears as its own nested sequence of
    helper items, intended to map to a tree-style widget such
    as a list view item. The data parameter must be a 2-tuple,
    with the first item being this item's strings (may be more
    than one because of columns), and the second item being a
    sequence of child data 2-tuples. The sequence of column
    strings is also exposed as the cols attribute.
    """
    
    colsclass = None
    
    def __init__(self, parent, index, data=None):
        self._parent = parent
        if data is not None:
            cols, children = data
        else:
            cols = children = None
        if not hasattr(self.__class__, 'item_class'):
            self.__class__.item_class = self.__class__
        if (self.colsclass is not None) and not hasattr(self, 'cols'):
            self.cols = self.colsclass(self, cols)
        self.children = children
        _PHelperItem.__init__(self, index)  # data must be None (it's in cols)
        _PTreeHelper.__init__(self, None, children)  # labels must be None
    
    def _value(self):
        # The item should appear as a sequence of 2-tuples
        return (self.cols, self.children)



#!/usr/bin/env python3
"""
Module WX.IMAGEVIEW -- Python wxWidgets Image View Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for image view widgets.
"""

import wx
from wx.lib import statbmp

from plib.ui.defs import *
from plib.ui.base.imageview import PImageViewBase

from .app import PWxWidget


WX_IMAGE_FORMATS = [
    "bmp",
    "png",
    "jpg",
    "jpeg",
    "gif",
    "pcx",
    "pnm",
    "tif",
    "tiff",
    "tga",
    "iff",
    "xpm",
    "ico",
    "cur",
    "ani",
]


class PImageView(PWxWidget, wx.ScrolledWindow, PImageViewBase):
    
    _align = False  # used by panel to determine placement
    
    child = None
    
    def __init__(self, manager, parent, filename=None, geometry=None):
        wx.ScrolledWindow.__init__(self, parent)
        PImageViewBase.__init__(self, manager, parent, filename=filename, geometry=geometry)
    
    @staticmethod
    def supported_formats():
        # TODO: is there a way to query wx for this?
        return sorted(WX_IMAGE_FORMATS)
    
    def load_from_file(self, filename):
        bitmap = wx.Bitmap(filename)
        width, height = bitmap.GetWidth(), bitmap.GetHeight()
        self.child = statbmp.GenStaticBitmap(self, -1, bitmap)
        self.SetScrollbars(20, 20, width // 20, height // 20)
    
    def get_image_size(self):
        bitmap = self.child.GetBitmap()
        return bitmap.GetWidth(), bitmap.GetHeight()
    
    def zoom_to(self, width, height):
        bitmap = self.child.GetBitmap()
        w, h = bitmap.GetWidth(), bitmap.GetHeight()
        wfactor, hfactor = (width / w), (height / h)
        if wfactor < hfactor:
            height = round(h * wfactor)
        elif hfactor < wfactor:
            width = round(w * hfactor)
        image = bitmap.ConvertToImage()
        image.Rescale(width, height)
        self.child.SetBitmap(image.ConvertToBitmap())
    
    def rotate_90(self, clockwise):
        bitmap = self.child.GetBitmap()
        image = bitmap.ConvertToImage()
        self.child.SetBitmap(image.Rotate90(clockwise).ConvertToBitmap())



#!/usr/bin/env python3
"""
Module PYSIDE2.GROUPBOX -- Python PySide 2 Group Box Widgets
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for group box widgets.
"""

from PySide2 import QtWidgets as qt

from plib.ui.base.groupbox import PGroupBoxBase

from .app import PQtWidget


class PGroupBox(PQtWidget, qt.QGroupBox, PGroupBoxBase):
    
    def __init__(self, manager, parent, caption, contents=None,
                 geometry=None, margin=None, spacing=None):
        
        qt.QGroupBox.__init__(self, parent)
        self.setSizePolicy(qt.QSizePolicy.MinimumExpanding,
                           qt.QSizePolicy.Fixed)
        self._vlayout = qt.QVBoxLayout()
        PGroupBoxBase.__init__(self, manager, parent, caption, contents,
                               geometry=geometry, margin=margin, spacing=spacing)
    
    def set_caption(self, caption):
        self.setTitle(caption)
    
    def set_margin(self, margin):
        self._vlayout.setContentsMargins(margin, margin, margin, margin)
    
    def set_spacing(self, spacing):
        self._vlayout.setSpacing(spacing)
    
    def add_control(self, control):
        self._vlayout.addWidget(control)
    
    def do_layout(self):
        self._vlayout.addStretch(1)
        self.setLayout(self._vlayout)



#!/usr/bin/env python3
"""
Module WX.LABEL -- Python wxWidgets Text Label Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for text label widgets.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.label import PTextLabelBase

from .app import PWxWidget


class PTextLabel(PWxWidget, wx.StaticText, PTextLabelBase):
    
    _align = False  # used by panel to determine placement
    
    def __init__(self, manager, parent, text=None,
                 geometry=None, font=None):
        
        wx.StaticText.__init__(self, parent)
        PTextLabelBase.__init__(self, manager, parent, text,
                                geometry=geometry, font=font)
    
    fn_get_text = 'GetLabel'
    fn_set_text = 'SetLabel'



#!/usr/bin/env python3
"""
Module WX.DIALOG -- Python wxWidgets Dialog Widget
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the dialog widget.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.dialog import PDialogBase

from .app import PWxWidget


class PDialog(PWxWidget, wx.Dialog, PDialogBase):
    
    def __init__(self, manager, parent, caption, client=None,
                 accept_buttons=('ok', 'yes'), reject_buttons=('cancel', 'no')):
        
        wx.Dialog.__init__(self, parent)
        PDialogBase.__init__(self, manager, parent, caption, client,
                             accept_buttons=accept_buttons, reject_buttons=reject_buttons)
        self.handlers = []
    
    def set_caption(self, caption):
        self.SetTitle(caption)
    
    def set_size(self, width, height):
        self.SetClientSize(width, height)
    
    def do_display(self):
        self.Show(True)
    
    def connect_target(self, signal, target):
        if signal == SIGNAL_FINISHED:
            # FIXME: wx dialogs do not seem to fire the EVT_CLOSE
            # event when the docs say they should
            self.handlers.append(target)
        else:
            super(PDialog, self).connect_target(signal, target)
    
    def EndDialog(self, code):
        for handler in self.handlers:
            handler(code)
        self.Show(False)
        self.Destroy()
    
    def accept_changes(self):
        self.EndDialog(1)
    
    def reject_changes(self):
        self.EndDialog(0)



#!/usr/bin/env python3
"""
Module QT5.TABLE -- Python Qt 5 Table Objects
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for the table widgets.
"""

import functools

from PyQt5 import QtGui as qtg, QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.table import PTableLabelsBase, PTableBase

from .app import PQtSequenceMeta, PQtWidgetBase, align_map
from .header import PQtHeaderBase, PQtHeaderWidget


class PTableLabels(PQtHeaderBase, PTableLabelsBase):
    
    def _set_labels_from_list(self, label_list):
        self.table.setHorizontalHeaderLabels(label_list)
    
    def _header_item(self, index):
        return self.table.horizontalHeaderItem()
    
    def _header_view(self):
        return self.table.horizontalHeader()

    def _set_align(self, index, align):
        self.table.horizontalHeaderItem(index).setTextAlignment(align_map[align])
        # Table cells will align themselves when added


class PTable(qt.QTableWidget, PQtHeaderWidget, PQtWidgetBase, PTableBase,
             metaclass=PQtSequenceMeta):
    
    widget_class = qt.QTableWidget
    
    labels_class = PTableLabels
    
    event_signals = (SIGNAL_CELLCHANGED,)
    
    def __init__(self, manager, parent, labels=None, data=None,
                 font=None, header_font=None):
        
        qt.QTableWidget.__init__(self, parent)
        self.setSortingEnabled(False)
        # Used by ugly hack in default fg and bk color methods, below
        pal = self.palette()
        self._def_fgcolor = pal.color(self.foregroundRole())
        self._def_bkcolor = pal.color(self.backgroundRole())
        # Used by ugly hack in setup_notify, below
        self._setting_colors = False
        
        # This will initialize data (if any)
        PTableBase.__init__(self, manager, parent, labels=labels, data=data,
                            font=font, header_font=header_font)
        
        # enable signals after initializing data so they don't get
        # fired on initialization
        self.setup_notify(SIGNAL_TABLECHANGED, self._tablechanged)
    
    def _tablechanged(self, row, col):
        # Filter out the signals we're not interested in
        if (row == self.current_row()) and (col == self.current_col()):
            self.do_notify(SIGNAL_CELLCHANGED, row, col)
    
    def wrap_target(self, signal, target):
        # Hack to mask out table changed events fired when cell colors are
        # changed (the Qt 4 API says this shouldn't happen but it does,
        # go figure)
        if signal == SIGNAL_TABLECHANGED:
            @functools.wraps(target)
            def _wrapper(row, col):
                if not self._setting_colors:
                    target(row, col)
            return _wrapper
        return target
    
    def _get_item(self, row, col):
        result = self.item(row, col)
        if not isinstance(result, qt.QTableWidgetItem):
            result = qt.QTableWidgetItem()
            self.setItem(row, col, result)
        return result
    
    def _get_cell(self, row, col):
        # Need str conversion here since widgets return QStrings
        return str(self._get_item(row, col).text())
    
    def _set_cell(self, row, col, value):
        item = self._get_item(row, col)  # this will fire table changed if row added
        item.setText(str(value))
        # FIXME: it would be nice if this could be done once instead of
        # per item
        item.setTextAlignment(self.horizontalHeaderItem(col).textAlignment())
    
    def header_object(self):
        return self.horizontalHeader()
    
    def rowcount(self):
        return self.rowCount()
    
    def colcount(self):
        return self.columnCount()
    
    def set_colcount(self, count):
        self.setColumnCount(count)
    
    def current_row(self):
        return self.currentRow()
    
    def current_col(self):
        return self.currentColumn()
    
    def _insert_row(self, index):
        self.insertRow(index)
    
    def _remove_row(self, index):
        self.removeRow(index)
    
    def topmargin(self):
        return 0  # self.topMargin()
    
    def leftmargin(self):
        return 0  # self.leftMargin()
    
    def rowheight(self, row):
        return self.rowHeight(row)
    
    def colwidth(self, col):
        return self.columnWidth(col)
    
    def default_fgcolor(self):
        return self._def_fgcolor
    
    def default_bkcolor(self):
        return self._def_bkcolor
    
    def set_text_fgcolor(self, row, col, color):
        item = self._get_item(row, col)
        self._setting_colors = True
        item.setForeground(qtg.QBrush(self._mapped_color(color)))
        self._setting_colors = False
    
    def set_cell_bkcolor(self, row, col, color):
        item = self._get_item(row, col)
        self._setting_colors = True
        item.setBackground(qtg.QBrush(self._mapped_color(color)))
        self._setting_colors = False



#!/usr/bin/env python3
"""
Module LABEL -- UI Text Label Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from .app import PWidgetBase


class PTextLabelBase(PWidgetBase):
    """Base class for widget that displays a single line of text.
    """
    
    def __init__(self, manager, parent, caption="",
                 geometry=None, font=None):
        
        PWidgetBase.__init__(self, manager, parent,
                             geometry=geometry, font=font)
        self.caption = caption
    
    fn_get_text = None
    fn_set_text = None
    
    def get_text(self):
        return getattr(self, self.fn_get_text)()
    
    def set_text(self, value):
        getattr(self, self.fn_set_text)(value)
    
    caption = property(get_text, set_text)



#!/usr/bin/env python3
"""
GRAPHICS.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A basic graphics demo using a container.
"""

import sys
import os

from plib.ui.toolkits import set_toolkit, ToolkitError

# We have to do this before any other plib.ui modules are imported,
# to avoid triggering the toolkit auto-detection machinery

try:
    set_toolkit('pyside2')
except ToolkitError:
    try:
        set_toolkit('qt5')
    except ToolkitError:
        print("Can't find pyside2 or qt5 toolkit.")
        sys.exit(1)

from plib.ui import __version__
from plib.ui.app import PApplication
from plib.ui.widgets import *  # this also imports get_toolkit


def make_graphic(parent, *filenames):
    print("Setting up graphic")
    toolkit = get_toolkit()
    if toolkit == 'pyside2':
        from PySide2 import QtGui, QtWidgets
    elif toolkit == 'qt5':
        from PyQt5 import QtGui, QtWidgets
    
    scene = QtWidgets.QGraphicsScene(parent)
    
    background = None
    if len(filenames) > 0:
        print("Loading background from", filenames[0])
        background = scene.addPixmap(QtGui.QPixmap(filenames[0]))
        background.setScale(0.2)
    
    sprites = []
    if len(filenames) > 1:
        for filename in filenames[1:]:
            print("Loading sprite from", filenames[1])
            sprite = scene.addPixmap(QtGui.QPixmap(filename))
            sprite.setScale(0.05)
            sprite.setPos(510.0 - 37.5, 255.0 - 26.25)
            sprites.append(sprite)
    
    graphic = QtWidgets.QGraphicsView(scene, parent)
    return graphic, background, sprites


class GraphicsDisplay(PApplication):
    
    about_data = {
        'name': "GraphicsDisplay",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "Graphics Display Tester",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net"
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    main_title = "Graphics Display Tester"
    
    main_widget = frame(ALIGN_JUST, LAYOUT_VERTICAL, [
        container('display'),
        panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
            padding(),
            button('left', "<"),
            combo('sprite'),
            num_edit('move_by'),
            button('right', ">"),
            padding(),
        ]),
        panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
            padding(),
            action_button(ACTION_ABOUT),
            action_button(ACTION_ABOUT_TOOLKIT),
            action_button(ACTION_EXIT),
        ]),
    ])
    
    # TODO: file open/save needed?
    menu_actions = [
        (MENU_FILE, (ACTION_FILE_OPEN, ACTION_FILE_SAVE, ACTION_EXIT,)),
        (MENU_HELP, (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT,)),
    ]
    
    toolbar_actions = [
        (ACTION_FILE_OPEN, ACTION_FILE_SAVE,),
        (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT, ACTION_EXIT,),
    ]
    
    graphic = None
    background = None
    sprites = ()
    
    move_unit = 1.0
    
    def after_create(self):
        if len(sys.argv) > 1:
            filenames = sys.argv[1:]
            self.graphic, self.background, self.sprites = make_graphic(self.container_display, *filenames)
            self.container_display.add_child(self.graphic)
            if len(filenames) > 1:
                self.combo_sprite.extend(os.path.basename(filename) for filename in filenames[1:])
    
    def on_app_shown(self):
        self.main_window.center()  # do after widgets are constructed and sized
    
    def move_sprite(self, left_right, up_down):
        index = self.combo_sprite.current_index()
        if index > -1:
            move_by = self.edit_move_by.edit_value * self.move_unit
            if move_by:
                sprite = self.sprites[index]
                sprite.moveBy(move_by * left_right, move_by * up_down)
            else:
                self.message_box.error("Move Error", "Can't move by zero units")
        else:
            self.message_box.error("Move Error", "No sprite selected")
    
    def on_left_clicked(self):
        self.move_sprite(-1, 0)
    
    def on_right_clicked(self):
        self.move_sprite(1, 0)
    
    def on_about(self):
        self.about()
    
    def on_about_toolkit(self):
        self.about_toolkit()
    
    def on_exit(self):
        self.exit_app()


if __name__ == "__main__":
    GraphicsDisplay().run()



#!/usr/bin/env python3
"""
Module WX.FORM -- Python wxWidgets Panel Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the panel widgets.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.form import PPanelBase

from .app import PWxWidget


wx_styles = {
    PANEL_NONE: wx.NO_BORDER,
    PANEL_BOX: wx.SIMPLE_BORDER,
    PANEL_RAISED: wx.RAISED_BORDER,
    PANEL_SUNKEN: wx.SUNKEN_BORDER
}

wx_layouts = {
    LAYOUT_NONE: wx.HORIZONTAL,
    LAYOUT_HORIZONTAL: wx.HORIZONTAL,
    LAYOUT_VERTICAL: wx.VERTICAL
}


class PPanel(PWxWidget, wx.Panel, PPanelBase):
    
    _expand = True  # used by parent panel to determine placement
    
    def __init__(self, manager, parent, align=ALIGN_JUST, layout=LAYOUT_NONE, contents=None,
                 style=PANEL_NONE, margin=None, spacing=None):
        
        wx.Panel.__init__(self, parent, style=wx_styles[style])
        self._align = (align == ALIGN_JUST)
        self._margin = margin
        self._spacing = spacing
        self._haswidgets = False
        self._sizer = wx.BoxSizer(wx_layouts[layout])
        PPanelBase.__init__(self, manager, parent, align, layout, contents=contents,
                            style=style, margin=margin, spacing=spacing)
    
    def set_min_size(self, width, height):
        self._sizer.SetMinSize((width, height))
    
    def set_margin(self, margin):
        # Margin is dealt with differently in wxWidgets, see above
        pass
    
    def set_spacing(self, spacing):
        # Spacing is dealt with differently in wxWidgets, see above
        pass
    
    def add_widget(self, widget):
        # People say the wxWidgets layout model is simple and easy to use,
        # but comparing this cruft with the PySide/Qt code, I'm not so sure...
        
        if self._haswidgets:
            if self._spacing is not None:
                self._sizer.AddSpacer(self._spacing)
        elif self._margin is not None:
            self._sizer.AddSpacer(self._margin)
        
        if getattr(widget, '_align', None):
            proportion = 1
        else:
            proportion = 0
        
        if getattr(widget, '_expand', None):
            flag = wx.EXPAND
        else:
            flag = 0
        
        exp = getattr(widget, '_expand_horiz', None)
        if self._sizer.GetOrientation() == wx.VERTICAL:
            horiz_pad = self._margin or 0
            vert_pad = None
            if exp:
                flag |= wx.EXPAND
        else:
            horiz_pad = None
            vert_pad = self._margin or 0
            if exp:
                proportion = 1
        
        if horiz_pad is not None:
            flag |= wx.LEFT | wx.RIGHT
            border = horiz_pad
        elif vert_pad is not None:
            flag |= wx.TOP | wx.BOTTOM
            border = vert_pad
        else:
            border = 0
        
        # At last we can actually do what we came for...
        self._sizer.Add(widget, proportion, flag, border)
        if not self._haswidgets:
            self._haswidgets = True
    
    def do_layout(self):
        if self._margin is not None:
            self._sizer.AddSpacer(self._margin)
        self.SetSizerAndFit(self._sizer)



#!/usr/bin/env python3
"""
Module WX.TABWIDGET -- Python wxWidgets Tab Widget
Sub-Package UI.TOOLKITS.QT of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the tab widget.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.tabwidget import PTabWidgetBase

from .app import PWxSequenceWidget


class PTabWidget(PWxSequenceWidget, wx.Notebook, PTabWidgetBase):
    
    _align = True  # used by panel to determine placement
    _expand = True
    
    def __init__(self, manager, parent, tabs=None):
        wx.Notebook.__init__(self, parent)
        PTabWidgetBase.__init__(self, manager, parent, tabs=tabs)
    
    def tab_count(self):
        return self.GetPageCount()
    
    def get_tab_title(self, index):
        return self.GetPageText(index)
    
    def set_tab_title(self, index, title):
        self.SetPageText(index, title)
    
    def tab_at(self, index):
        return self.GetPage(index)
    
    def add_tab(self, index, title, widget):
        if index == self.__len__():
            self.AddPage(widget, title)
        else:
            self.InsertPage(index, widget, title)
    
    def del_tab(self, index):
        self.RemovePage(index)
    
    def current_index(self):
        return self.GetSelection()
    
    def set_current_index(self, index):
        self.SetSelection(index)



#!/usr/bin/env python3
"""
Module COMBO -- UI Combo Box Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import collections

from plib.stdlib.builtins import inverted
from plib.stdlib.coll import baselist, SortMixin

from plib.ui.defs import *

from .app import PWidgetBase


class PComboBoxBase(PWidgetBase, baselist):
    """Combo box that looks like a list of strings.
    
    (Note: currently selection is limited to items added to the
    combo box programmatically; the user cannot edit in the edit
    control and add new items to the pick list.)
    """
    
    signals = (
        SIGNAL_SELECTED,
    )
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None):
        
        PWidgetBase.__init__(self, manager, parent,
                             geometry=geometry)
        baselist.__init__(self, items)
        self.complete_init(items, value)
    
    def complete_init(self, items, value):
        if value:
            self.set_current_text(value)
        elif items:
            self.set_current_index(0)
    
    def current_text(self):
        return self[self.current_index()]
    
    get_current_text = current_text
    
    def set_current_text(self, text):
        self.set_current_index(self.index(text))
    
    def current_index(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_current_index(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PNumComboBoxBase(PComboBoxBase):
    """Combo box that maps strings to integer values.
    """
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None):
        
        self.text_map = dict(items)  # value -> text
        self.value_map = inverted(self.text_map)  # text -> value
        str_items = [item[1] for item in items]
        PComboBoxBase.__init__(self, manager, parent, items=str_items,
                               geometry=geometry)
        if value:
            self.set_current_value(value)  # value will be an int
    
    def current_value(self):
        return self.value_map[self.current_text()]
    
    get_current_value = current_value
    
    def set_current_value(self, value):
        self.set_current_text(self.text_map[value])


class PSortedComboBoxBase(SortMixin, PComboBoxBase):
    """Combo box that automatically sorts its items.
    """
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, key=None):
        
        PComboBoxBase.__init__(self, manager, parent, geometry=geometry)  # don't pass items here
        self._init_seq(items, key)  # sort items and add them here
        self.complete_init(items, value)  # have to do this after data is added



#!/usr/bin/env python3
"""
Module GROUPBOX -- UI Group Box Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.widgets import widgets_from_contents

from .app import PWidgetBase


class PGroupBoxBase(PWidgetBase):
    """Base class for group box widget.
    """
    
    controls = None
    
    def __init__(self, manager, parent, caption, contents=None,
                 geometry=None, margin=None, spacing=None):
        
        PWidgetBase.__init__(self, manager, parent,
                             geometry=geometry)
        
        self.set_caption(caption)
        if (margin is not None):
            self.set_margin(margin)
        if (spacing is not None):
            self.set_spacing(spacing)
        self.init_controls(contents)
        self.do_layout()
    
    def set_caption(self, caption):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_margin(self, margin):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_spacing(self, spacing):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def init_controls(self, contents):
        if self.controls is None:
            self.controls = []
        self.controls.extend(widgets_from_contents(self.manager, self, contents, self.add_control))
    
    def add_control(self, control):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def do_layout(self):
        # Derived classes should override to do any
        # necessary creation of physical layout objects
        # after all child controls are created.
        pass



#!/usr/bin/env python3
"""
Module SOCKET -- UI Socket Handler Classes
Sub-Package UI.SOCKET of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""


class PSocketNotifierBase(object):
    
    def __init__(self, obj, notify_type, select_fn, notify_fn):
        self._obj = obj
        self._notify_type = notify_type
        self.select_fn = select_fn
        self.notify_fn = notify_fn
    
    def set_enabled(self, enable):
        """Enable or disable this specific notifier.
        """
        raise NotImplementedError
    
    @classmethod
    def start(cls):
        """Tell the notifier event loop to start processing socket events.
        """
        pass
    
    @classmethod
    def done(cls):
        """Tell the notifier event loop to stop processing socket events.
        """
        pass



#!/usr/bin/env python3
"""
Module PYSIDE2.FORM -- Python PySide 2 Panel Objects
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for the panel widgets.
"""

from PySide2 import QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.form import PPanelBase

from .app import PQtWidget


panel_map = {
    PANEL_NONE: qt.QFrame.NoFrame | qt.QFrame.Plain,
    PANEL_BOX: qt.QFrame.Box | qt.QFrame.Plain,
    PANEL_RAISED: qt.QFrame.Panel | qt.QFrame.Raised,
    PANEL_SUNKEN: qt.QFrame.Panel | qt.QFrame.Sunken
}


def size_policy(align):
    horiz = vert = qt.QSizePolicy.MinimumExpanding
    if align in (ALIGN_LEFT, ALIGN_RIGHT):
        horiz = qt.QSizePolicy.Fixed
    elif align in (ALIGN_TOP, ALIGN_BOTTOM):
        vert = qt.QSizePolicy.Fixed
    return horiz, vert


class PPanel(PQtWidget, qt.QFrame, PPanelBase):
    
    def __init__(self, manager, parent, align=ALIGN_JUST, layout=LAYOUT_NONE, contents=None,
                 style=PANEL_NONE, margin=None, spacing=None):
        
        qt.QFrame.__init__(self, parent)
        self.setFrameStyle(panel_map[style])
        self.setSizePolicy(*size_policy(align))
        if layout == LAYOUT_HORIZONTAL:
            klass = qt.QHBoxLayout
        else:
            klass = qt.QVBoxLayout
        self._playout = klass()
        PPanelBase.__init__(self, manager, parent, align, layout, contents=contents,
                            style=style, margin=margin, spacing=spacing)
        # Qt 4/5 defaults don't seem to be the same as Qt 3, so compensate
        if margin is None:
            self.set_margin(0)
        if spacing is None:
            self.set_spacing(0)
    
    def set_box_width(self, width):
        self.setLineWidth(width)
    
    def set_margin(self, margin):
        self._playout.setContentsMargins(margin, margin, margin, margin)
    
    def set_spacing(self, spacing):
        self._playout.setSpacing(spacing)
    
    def add_widget(self, widget):
        self._playout.addWidget(widget)
    
    def do_layout(self):
        #self._playout.addStretch(1)
        self.setLayout(self._playout)



#!/usr/bin/env python3
"""
Module PREFS -- UI Preferences Dialog
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import collections

from plib.ui.defs import *
from plib.ui.dialogs import DialogRunner
from plib.ui.widgets import *


class PPreferences(DialogRunner):
    """Manage editing and saving of program preferences.
    """
    
    default_dialog_width = 480
    default_group_style = PANEL_SUNKEN
    default_label_bold = False
    default_group_label_bold = True
    
    def __init__(self, app, caption):
        DialogRunner.__init__(self, app, caption)
        self.inifile, self.section_type, self.field_map = app.prefs_data
        self.dialog_enums = getattr(app, 'prefs_dialog_enums', {})
        self.dialog_groups = getattr(app, 'prefs_dialog_groups', {})
        self.dialog_width = getattr(app, 'prefs_dialog_width', self.default_dialog_width)
        self.group_style = getattr(app, 'prefs_group_style', self.default_group_style)
        self.label_bold = getattr(app, 'prefs_label_bold', self.default_label_bold)
        self.group_label_bold = getattr(app, 'prefs_group_label_bold', self.default_group_label_bold)
        self.dialog_client = self.get_client_spec()
    
    def unpack_option(self, sname, opt):
        attrname = getter = setter = None
        if len(opt) > 4:
            oname, otype, odefault, getter, setter = opt
        elif len(opt) > 3:
            oname, otype, odefault, attrname = opt
        elif len(opt) > 2:
            oname, otype, odefault = opt
        else:
            oname, odefault = opt
            otype = type(odefault)
        if attrname is None:
            attrname = '{}_{}'.format(sname, oname)
        return otype, odefault, attrname
    
    def get_widget(self, attrname, otype, odefault, caption, geometry=None):
        if otype is bool:
            # No enum possible for checkbox, no separate label
            return checkbox(attrname, caption, odefault)
        
        enum = self.dialog_enums.get(attrname)
        control = (
            (num_combo if otype is int else combo)(attrname, enum, odefault)
            if enum else
            (num_edit if otype is int else edit)(attrname, odefault, geometry=geometry)
        )
        
        return labeled(caption, control, label_bold=self.label_bold)
    
    def get_client_spec(self):
        geometry = (None, None, self.dialog_width, None)
        
        contents = []
        for sname, opts in self.inifile._optionlist:
            specs = collections.OrderedDict()
            for opt in opts:
                otype, odefault, attrname = self.unpack_option(sname, opt)
                caption = self.field_map.get(attrname, "<Unknown Setting>")
                specs[attrname] = (otype, odefault, caption)
            
            groups = self.dialog_groups.get(sname)
            if not groups:
                groups = [(key,) for key in specs]
            controls = []
            for group in groups:
                if len(group) == 1:
                    attrname = group[0]
                    otype, odefault, caption = specs[attrname]
                    widget = self.get_widget(attrname, otype, odefault, caption, geometry)
                else:
                    widget = panel(ALIGN_JUST, LAYOUT_HORIZONTAL, [
                        self.get_widget(attrname, *specs[attrname])
                        for attrname in group
                    ])
                controls.append(widget)
            
            group_caption = self.field_map.get(sname)
            contents.append((group_caption, controls))
        
        if self.section_type == SECTION_TAB:
            dialog_controls = tabwidget('prefsdialog', [
                (caption, tab(ALIGN_JUST, LAYOUT_VERTICAL, controls + [padding()]))
                for caption, controls in contents
            ])
        elif self.section_type == SECTION_GROUPBOX:
            dialog_controls = panel(ALIGN_JUST, LAYOUT_VERTICAL, [
                labelbox(caption, controls, align=ALIGN_TOP, style=self.group_style, label_bold=self.group_label_bold)
                for caption, controls in contents
            ])
        else:
            raise ValueError("Unknown dialog section type: {}".format(repr(section_type)))
        
        dialog_buttons = panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
            padding(),
            action_button(ACTION_APPLY),
            action_button(ACTION_OK),
            action_button(ACTION_CANCEL),
        ])
        
        return frame(ALIGN_JUST, LAYOUT_VERTICAL, [
            dialog_controls,
            dialog_buttons,
        ])
    
    def iter_vars(self):
        for sname, opts in self.inifile._optionlist:
            for opt in opts:
                otype, odefault, attrname = self.unpack_option(sname, opt)
                yield otype, odefault, attrname
    
    def control_by_name(self, otype, attrname):
        return getattr(self, '{}_{}'.format(
            (
                'checkbox' if otype is bool else
                'combo' if attrname in self.dialog_enums else
                'edit'
            ),
            attrname
        ))
    
    def control_method(self, otype, attrname):
        return (
            'checked' if otype is bool else
            ('current_{}' if attrname in self.dialog_enums else "{}").format(
                'value' if otype is int else 'text'
            )
        )
    
    def control_set(self, otype, attrname, value):
        control = self.control_by_name(otype, attrname)
        method = 'set_{}'.format(self.control_method(otype, attrname))
        getattr(control, method)(value)
    
    def populate(self):
        for otype, _, attrname in self.iter_vars():
            self.control_set(otype, attrname, getattr(self.inifile, attrname))
    
    def control_get(self, otype, attrname):
        control = self.control_by_name(otype, attrname)
        method = 'get_{}'.format(self.control_method(otype, attrname))
        return getattr(control, method)()
    
    def get_result(self):
        for otype, _, attrname in self.iter_vars():
            setattr(self.inifile, attrname, self.control_get(otype, attrname))
        self.inifile.writeini()
    
    def on_apply(self):
        self.get_result()
    
    def dialog_done(self, accepted):
        # We don't need to pass a callback in the constructor since we can
        # do all the work here, we just override this method directly
        print("Dialog accepted:", repr(accepted))
        if accepted:
            self.get_result()



#!/usr/bin/env python3
"""
Module WX.MAINWIN -- Python wxWidgets Main Window Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI main window objects.
"""

import wx
from wx.lib.evtmgr import eventManager

from plib.ui.defs import *
from plib.ui.base.mainwin import PMenuBase, PToolBarBase, PStatusLabelBase, PStatusBarBase, PActionBase, PMainWindowBase

from .app import PWxSignal, PWxActionMixin, PWxWidget, PTopWindow, scaled_bitmap, stock_ids, parse_accel


class PWxPopup(wx.Menu):
    """Customized wxWidgets popup menu class.
    """
    
    def __init__(self, main_window):
        self.main_window = main_window
        astyle = 0
        wx.Menu.__init__(self, "", astyle)


class PWxMenu(PWxWidget, wx.MenuBar, PMenuBase):
    """Customized wxWidgets menu class.
    """
    
    popup_class = PWxPopup
    
    def __init__(self, main_window):
        astyle = 0
        wx.MenuBar.__init__(self, astyle)
        PMenuBase.__init__(self, main_window)
    
    def do_add_popup(self, title, popup):
        self.Append(popup, title)
    
    def add_popup_action(self, action, popup):
        # FIXME: the stock menu ids don't enable standard images
        if action._id > wx.ID_HIGHEST:
            item = wx.MenuItem(popup, action._id, action.menustr, action.statusbarstr)
        else:
            item = wx.MenuItem(popup, action._id)
        item.SetBitmap(scaled_bitmap(action.image, action.menufactor))
        accel = action.accelstr
        if accel is not None:
            flags, keycode = parse_accel(accel)
            item.SetAccel(wx.AcceleratorEntry(flags, keycode, cmd=action._id, item=item))
        popup.Append(item)


class PWxToolBar(PWxWidget, wx.ToolBar, PToolBarBase):
    """Customized wxWidgets toolbar class.
    """
    
    def __init__(self, main_window):
        astyle = 0
        if main_window.app.show_labels:
            astyle = astyle | wx.TB_TEXT
        wx.ToolBar.__init__(self, main_window, style=astyle)
        PToolBarBase.__init__(self, main_window)
    
    def add_action(self, action):
        s = action.toolbarstr if self.main_window.app.show_labels else ""
        img = scaled_bitmap(action.image, action.toolbarfactor)
        # FIXME: The stock toolbar ids don't enable standard captions, images,
        # tooltips, or status bar text
        toolid = action._id
        self.AddTool(toolid, s, img)
        self.SetToolShortHelp(toolid, action.toolbarstr)
        self.SetToolLongHelp(toolid, action.statusbarstr)
    
    def add_separator(self):
        self.AddSeparator()


class PStatusLabel(PStatusLabelBase):
    
    def init_status_label(self):
        self.actual_widget = self
        self.status_bar = self._parent
        self.field_index = -1
    
    fn_get_text = 'GetText'
    fn_set_text = 'SetText'
    
    def GetText(self):
        return self.status_bar.GetStatusText(self.field_index)
    
    def SetText(self, value):
        self.status_bar.SetStatusText(value, self.field_index)
        if self.status_bar._initialized:
            self.status_bar.SetStatusWidths(self.status_bar.label_widths)


status_label_styles = {
    PANEL_NONE: wx.SB_NORMAL,
    PANEL_BOX: wx.SB_NORMAL,  # TODO: is there any way to get a box border in wx?
    PANEL_RAISED: wx.SB_RAISED,
    PANEL_SUNKEN: wx.SB_SUNKEN,
}


class PWxStatusBar(PWxWidget, wx.StatusBar, PStatusBarBase):
    
    def __init__(self, main_window):
        self._initialized = False
        wx.StatusBar.__init__(self, main_window)
        PStatusBarBase.__init__(self, main_window)
    
    def add_widget(self, widget, expand=False, custom=True):
        self.labels.append(widget)
    
    char_width = 12  # TODO: is there a way to get this info from wx?
    
    @property
    def label_widths(self):
        return [-1] + [self.char_width * len(label.caption) for label in self.labels]
    
    @property
    def label_styles(self):
        label_styles = [PANEL_NONE] + [label._style for label in self.labels]
        return [status_label_styles[style] for style in label_styles]
    
    def create_widgets(self):
        self.labels = []
        super(PWxStatusBar, self).create_widgets()
        
        self.SetFieldsCount(1 + len(self.labels))
        self.SetStatusStyles(self.label_styles)
        for i, label in enumerate(self.labels):
            label.field_index = i + 1  # so GetText and SetText will access the right field
            label.caption = label._caption
        self.SetStatusWidths(self.label_widths)
        
        self._initialized = True  # so individual labels will trigger a width update if their caption changes
    
    def get_text(self):
        return self.GetStatusText(0)
    
    def set_text(self, value):
        self.SetStatusText(value, 0)


class PWxAction(PWxActionMixin, PWxSignal, PActionBase):
    """Customized wxWidgets action class.
    """
    
    def __init__(self, key, main_window):
        PActionBase.__init__(self, key, main_window)
        
        self.menufactor = 0.5
        if main_window.app.large_icons:
            self.toolbarfactor = None
        else:
            self.toolbarfactor = 0.6875
        self.menustr = self.get_menu_str(key)
        self.toolbarstr = self.get_toolbar_str(key)
        self.statusbarstr = self.get_statusbar_str(key)
        self.accelstr = self.get_accel_str(key)
        self._id = stock_ids.get(key, wx.ID_HIGHEST + key)
        self.image = self.get_icon(key)
    
    def wx_register_event(self, event, handler):
        # The usual eventManager.Register doesn't seem to work here; note
        # that this method means only a single handler can be tied to
        # action events (the latest one bound)
        self.main_window.Bind(event, handler, id=self._id)
    
    def enable(self, enabled):
        menu = self.main_window.menu
        toolbar = self.main_window.toolbar
        if menu is not None:
            menu.Enable(self._id, enabled)
        if toolbar is not None:
            toolbar.EnableTool(self._id, enabled)


class PMainWindow(PTopWindow, PMainWindowBase):
    """Customized wxWidgets main window class.
    """
    
    menu_class = PWxMenu
    toolbar_class = PWxToolBar
    statusbar_class = PWxStatusBar
    action_class = PWxAction
    
    action_signals = None
    
    def create_menu(self):
        menu = super(PMainWindow, self).create_menu()
        self.SetMenuBar(menu)
        return menu
    
    def create_toolbar(self):
        toolbar = super(PMainWindow, self).create_toolbar()
        self.SetToolBar(toolbar)
        return toolbar
    
    def create_statusbar(self):
        statusbar = super(PMainWindow, self).create_statusbar()
        self.SetStatusBar(statusbar)
        return statusbar
    
    def create_actions(self):
        actions = super(PMainWindow, self).create_actions()
        # Can't do this until after all actions are added
        if self.toolbar:
            self.toolbar.Realize()
        return actions



#!/usr/bin/env python3
"""
UI-TEXT.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A demo app to test the text display control and the
TextOutput file-like object wrapper for it.
"""

import sys
import os

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication
from plib.ui.output import PTextOutput
from plib.ui.widgets import *


class UITextOutputTester(PApplication):
    
    about_data = {
        'name': "UITextOutputTester",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "UI Text Output Test Demo",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net"
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    main_title = "UI Text Output Tester"
    
    main_size = SIZE_OFFSET
    width_offset = height_offset = 50
    main_placement = MOVE_CENTER
    
    main_widget = frame(ALIGN_JUST, LAYOUT_VERTICAL, [
        panel(ALIGN_TOP, LAYOUT_HORIZONTAL, [
            panel(ALIGN_JUST, LAYOUT_HORIZONTAL, [
                edit('text'),
            ]),
            panel(ALIGN_RIGHT, LAYOUT_HORIZONTAL, [
                button('add', "Add Text"),
                num_edit('num'),
                label('num', "times"),
                button('clear', "Clear Text"),
            ]),
        ]),
        text('output', font=("Courier New", 12)),
        panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
            labeled("Text Size:", label('size', "", font=(None, None, True)), LAYOUT_HORIZONTAL, ALIGN_LEFT, label_bold=False),
            label('unit', "Bytes"),
        ]),
    ])
    
    def after_create(self):
        self.outputfile = PTextOutput(self.text_output)
        self.edit_text.edit_text = "This is a line of text to be added to the text display."
        self.edit_num.edit_value = 100
        self.update_size()
    
    def update_size(self):
        s = self.outputfile.filesize
        print(s)
        self.label_size.caption = str(s)
    
    def add_output(self, s):
        self.outputfile.write(s)
        self.outputfile.write(os.linesep)
        self.outputfile.flush()
    
    def on_add_clicked(self):
        text = self.edit_text.edit_text
        if text:
            for _ in range(self.edit_num.edit_value):
                self.add_output(text)
            self.update_size()
    
    on_text_enter = on_add_clicked
    
    def on_clear_clicked(self):
        self.outputfile.clear()
        self.update_size()


if __name__ == "__main__":
    UITextOutputTester().run()



#!/usr/bin/env python3
"""
Module PYSIDE2.DIALOG -- Python PySide 2 Dialog Widget
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for the dialog widget.
"""

from PySide2 import QtWidgets as qt

from plib.ui.base.dialog import PDialogBase

from .app import PQtWidget


class PDialog(PQtWidget, qt.QDialog, PDialogBase):
    
    def __init__(self, manager, parent, caption, client=None,
                 accept_buttons=('ok', 'yes'), reject_buttons=('cancel', 'no')):
        
        qt.QDialog.__init__(self, parent)
        PDialogBase.__init__(self, manager, parent, caption, client,
                             accept_buttons=accept_buttons, reject_buttons=reject_buttons)
    
    def set_caption(self, caption):
        self.setWindowTitle(caption)
    
    def do_display(self):
        self.show()
    
    def accept_changes(self):
        self.accept()
    
    def reject_changes(self):
        self.reject()
